/*

		Betriebssystemunabhngige Routinen

*/


//------------------------------------------------------------------
//              W I N D O W S   9 x
//------------------------------------------------------------------
#ifdef CAN_DRIVER_FOR_WIN9X

#define WANTVXDWRAPS

#include <basedef.h>
#include <vmm.h>
#include <debug.h>
#include <vmmreg.h>
#include <vcomm.h>
#include <vpicd.h>
#include <vtd.h>
#include <vwin32.h>
#include <winerror.h>
#include <shell.h>

#include "canpar.h"
#include "utreiber.h"
#include "win9x\treiber.h"

#include <vxdwraps.h>
#include <configmg.h>
#include <regstr.h>

#define OPERATING_SYSTEM OS_WINDOWS9x

#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
/*

ioctl16 fehlt
// folgende EIntrge beim Laden des Treibers initialisieren
lpDIOCParms->pelican = 0;
lpDIOCParms->eff = 0;
// auerdem folgendes in die CANPAR-SDtruktur zustzlich
unsigned long canpar_baudrate=20000;
unsigned long canpar_access_code=0xffffffff;
unsigned long canpar_access_mask=0xffffffff;
*/

#endif




//------------------------------------------------------------------
//              W I N D O W S   N T 4.0
//------------------------------------------------------------------
#ifdef CAN_DRIVER_FOR_NT4
// Wenn diese Datei unter NT4 compiliert wird, mssen alle
// Dateien relativ zu \canpar\nt angegeben werden
#include "..\canpar.h"
#include "..\utreiber.h"

#include "ntddk.h"
#include "parallel.h"
#include "treiber.h"
#include "log.h"

#define OPERATING_SYSTEM OS_WINDOWSNT4

#endif


//------------------------------------------------------------------
//              L I N U X   K E R N E L  2 . 2 . 1 6  (SuSE 7.0)
//              L I N U X   K E R N E L  2 . 2 . 1 8  (SuSE 7.1)
//              L I N U X   K E R N E L  2 . 4 . 0    (SuSE 7.1)
//------------------------------------------------------------------
#ifdef CAN_DRIVER_FOR_LINUX
#include <linux/module.h>
#include <linux/parport.h>

#include "canpar.h"
#include "utreiber.h"
#include "linux.2.2/treiber.h"

#define OPERATING_SYSTEM OS_LINUX

#endif

//------------------------------------------------------------------
//              K O D E
//------------------------------------------------------------------
static const char zeichen_toupper[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
static const char zeichen_tolower[] = "abcdefghijklmnopqrstuvwxyz";

char mytoupper(char data)
{
	int p;
	if ((data >= 'a') && (data <= 'z'))
	{
		p = data-'a';
		data = zeichen_toupper[p];
	}
	return data;
}

char mytolower(char data)
{
	int p;
	if ((data >= 'A') && (data <= 'Z'))
	{
		p = data-'A';
		data = zeichen_tolower[p];
	}
	return data;
}

void wr_data(PCANDEVICE lpDIOCParms, unsigned char adr,unsigned char wert) // Wert in Register des 82C200 an Adresse adr schreiben
{
	switch(lpDIOCParms->portTyp)
	{
	default:
	case CANCARD_ELEKTOR:
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport & ~0x20);   // Datenport auf Ausgang schalten
		out2Data(lpDIOCParms,adr);                                  // Adresse auf Datenport
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport | 0x08);      // ALE=0
		out2Data(lpDIOCParms,wert);                                 // Wert auf Datenport
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport | 0x02);      // WR\=0
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport & ~0x02);     // WR\=1
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport & ~0x08);     // ALE=1
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport | 0x20);      // Datenport auf Eingang
		break;
	case CANCARD_CAN200:
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport & ~0x20);   // Datenport auf Ausgang schalten
		out2Data(lpDIOCParms,adr);                                  // Adresse auf Datenport
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport & ~0x08);      // ALE=0
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport | 0x01);     // RD/WR\=0
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport | 0x02);      // E=1
		out2Data(lpDIOCParms,wert);                                 // Wert auf Datenport
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport & ~0x02);     // E=0
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport | 0x28);     // Datenport auf Eingang + ALE=1
		break;
	}
}

unsigned char rd_data(PCANDEVICE lpDIOCParms, unsigned char adr) // Wert aus Register des 82C200 auslesen
{
	unsigned char bw;
	switch(lpDIOCParms->portTyp)
	{
	default:
	case CANCARD_ELEKTOR:
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport & ~0x20);    // Datenport auf Ausgang
		out2Data(lpDIOCParms,adr);                                       // Adresse auf Datenport
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport | 0x08);     // ALE=0
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport | 0x20);     // Datenport auf Eingang
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport & ~0x04);    // RD\=0
		bw=inData(lpDIOCParms);                                          // Datenport auslesen
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport | 0x04);     // RD\=1
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport & ~0x08);    // ALE=1;
		break;
	case CANCARD_CAN200:
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport & ~0x20);    // Datenport auf Ausgang
		out2Data(lpDIOCParms,adr);                                       // Adresse auf Datenport
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport & ~0x08);      // ALE=0
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport & ~0x01);     // RD/WR\=1
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport | 0x20);      // Datenport auf Eingang
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport | 0x02);      // E=1
		bw=inData(lpDIOCParms);                                          // Datenport auslesen
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport & ~0x02);     // E=0
		out2Controlport(lpDIOCParms,lpDIOCParms->statusport | 0x08);     // ALE=1
		break;
	}
	return(bw);
}

void SetToResetMode(PCANDEVICE lpDIOCParms)
{
	// ein Zugrifssschutz (Mutex) mu vorher gesetzt werden
	wr_data(lpDIOCParms, 0, rd_data(lpDIOCParms,0) | 0x01);
}
	
void SetToNormalMode(PCANDEVICE lpDIOCParms)
{
	// ein Zugrifssschutz (Mutex) mu vorher gesetzt werden
	wr_data(lpDIOCParms, 0, rd_data(lpDIOCParms,0) & 0x3e);
}

int iSetSJA1000Baudrate(PCANDEVICE lpDIOCParms)
{
	unsigned char tmp;
	unsigned long wbaud;
	long bpr,ts1,ts2;
	
	int  ts2array[]={2,3,4,5,1,6,7,0};
	int  x;
	long testwert,neutest=50;

	wbaud = lpDIOCParms->baudrate;
	if (wbaud < 5000) wbaud = 5000;
#ifdef DEMOVERSION
	if (wbaud > 400000) 
	{
		if (lpDIOCParms->admISRInfo && (lpDIOCParms->admISRAktuell >= 0))
		{
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			Event2UserProgramm(lpDIOCParms, EV2PRG_BAUDRATE_DEMO);
			CreateEvent2UserSpace(lpDIOCParms);
		}
		return 0;
	}
#else
	if (wbaud > 2000000) wbaud = 2000000;
#endif
	lpDIOCParms->baudrate = wbaud;
		
	testwert = lpDIOCParms->frequency/(wbaud*2);
	while(neutest > 0)
	{
			for(x=0;x<8;x++)
			{
				ts2 = ts2array[x];
				for(ts1=0x0f;ts1>=0;ts1--)
				{
					if ((testwert % (ts1+ts2+3)) == 0)
					{
						bpr  = testwert/(ts1+ts2+3)-1;
						if ((bpr >= 0) && (bpr < 0x40))
						{
							bpr |= 0x40;    // SJW.0 setzen
							tmp = (unsigned char)((ts2 << 4) + ts1);

							TRACE(("BPR0 %lx BPR1 %x",bpr,tmp));
							wr_can_rr(lpDIOCParms, 6,(unsigned char)bpr);
							wr_can_rr(lpDIOCParms, 7,tmp);
							return 1;
						}
					}
				}
			}
			testwert++;
			neutest--;
	}
	TRACE(("Kann Baudrate nicht einstellen"));
	lpDIOCParms->baudrate = 0; // Baudrate kann nicht eingestellt werden !
	return 0;
}

int SetSJA1000Baudrate(PCANDEVICE lpDIOCParms)
{
	if(!iSetSJA1000Baudrate(lpDIOCParms))
	{
		lpDIOCParms->baudrate = 20000;
		iSetSJA1000Baudrate(lpDIOCParms);
		return 0;
	}
	return 1;
}

void SetSJA1000AccessCode_Mask(PCANDEVICE lpDIOCParms)
{
	if (lpDIOCParms->pelican)
	{
		wr_can_rr(lpDIOCParms, 16,(unsigned char)(lpDIOCParms->access_code >> 24));         // ACC
		wr_can_rr(lpDIOCParms, 17,(unsigned char)(lpDIOCParms->access_code >> 16));         // ACC
		wr_can_rr(lpDIOCParms, 18,(unsigned char)(lpDIOCParms->access_code >> 8));         // ACC
		wr_can_rr(lpDIOCParms, 19,(unsigned char)lpDIOCParms->access_code);         // ACC
		wr_can_rr(lpDIOCParms, 20,(unsigned char)(lpDIOCParms->access_mask >> 24));         // Mask
		wr_can_rr(lpDIOCParms, 21,(unsigned char)(lpDIOCParms->access_mask >> 16));         // Mask
		wr_can_rr(lpDIOCParms, 22,(unsigned char)(lpDIOCParms->access_mask >> 8));         // Mask
		wr_can_rr(lpDIOCParms, 23,(unsigned char)lpDIOCParms->access_mask);         // Mask
	}
	else
	{
		wr_can_rr(lpDIOCParms, 4,(unsigned char)lpDIOCParms->access_code);	// Acceptance Code schreiben
		wr_can_rr(lpDIOCParms, 5,(unsigned char)lpDIOCParms->access_mask);	// Acceptance Mask schreiben
	}
}

unsigned long SJA1000_Commands(PCANDEVICE lpDIOCParms, unsigned char command)
{
		if(!lpDIOCParms->pelican) switch(command)   // BASIC CAN Modus
		{
		case CAN_COMMAND_STANDBY:
			wr_can(lpDIOCParms, 1,0x10);
			break;
		case CAN_COMMAND_STANDBY_OFF:
			wr_can(lpDIOCParms, 1,~0x10 & rd_can(lpDIOCParms,1));
			break;
		case CAN_COMMAND_ABORT_TRANSMISSION:
			IOProc12(lpDIOCParms);
			break;
		case CAN_COMMAND_CLEAR_OVERRUN:
			wr_can(lpDIOCParms, 1,0x08);
			break;
		default:
			return ERROR_INVALID;
		}
		else switch(command)						// Pelican Modus
		{
		case CAN_COMMAND_STANDBY:
			wr_can(lpDIOCParms, 0,0x10 | rd_can(lpDIOCParms,0));
			lpDIOCParms->standby=1;
			break;
		case CAN_COMMAND_STANDBY_OFF:
			wr_can(lpDIOCParms, 0,~0x10 & rd_can(lpDIOCParms,0));
			lpDIOCParms->standby=0;
			break;
		case CAN_COMMAND_ABORT_TRANSMISSION:
			IOProc12(lpDIOCParms);
			break;
		case CAN_COMMAND_CLEAR_OVERRUN:
			wr_can(lpDIOCParms, 1,0x08);
			break;
		case CAN_COMMAND_SELF_RECEPTION_REQUEST:
			wr_can(lpDIOCParms, 1,0x10);
			break;
		case CAN_COMMAND_LISTEN_ON:
			wr_can_rr(lpDIOCParms, 0,0x02 | rd_can(lpDIOCParms,0));
			break;
		case CAN_COMMAND_LISTEN_OFF:
			wr_can_rr(lpDIOCParms, 0,~0x02 & rd_can(lpDIOCParms,0));
			break;
		default:
			return ERROR_INVALID;
		}
		return NO_ERROR;
}

// Initialisierung des Can-Controllers SJA1000
int initSJA1000(PCANDEVICE lpDIOCParms)
{
	if (lpDIOCParms->pelican)
	{
		TRACE(("Init Pelican"));
		// CAN-Controller initialisieren
		wr_can_rr(lpDIOCParms, 1,0x08);         // Acceptance Filter Mode: single
		wr_can_rr(lpDIOCParms, 31, 0x80 | rd_can_rr(lpDIOCParms, 31)); // Pelican einschalten
		wr_can_rr(lpDIOCParms, 4,0xff);         // IRQ's
		if(lpDIOCParms->portTyp == CANCARD_CAN200)
		{
			wr_can_rr(lpDIOCParms, 8,0x1a);         // Output Control
		}
		else
		{
			wr_can_rr(lpDIOCParms, 8,0xc2);         // Output Control
		}
		wr_can_rr(lpDIOCParms, 13,96);          // error warning limit
		wr_can_rr(lpDIOCParms, 14,0x00);         // RX error counter
		wr_can_rr(lpDIOCParms, 15,0x00);         // TX error counter
		wr_can(lpDIOCParms, 0,0x00);         // Reset Request =0
		rd_can(lpDIOCParms, 3);
		
		TRACE(("InitSJA1000 OK"));
		SetSJA1000Baudrate(lpDIOCParms);
		SetSJA1000AccessCode_Mask(lpDIOCParms);
		return 1;
	}
	else
	{
		TRACE(("Init BasicCan"));
		// CAN-Controller initialisieren
		wr_can_rr(lpDIOCParms, 31, 0x7f & rd_can_rr(lpDIOCParms, 31)); // Pelican ausschalten
		wr_can_rr(lpDIOCParms, 0,0x3f);         // Reset Request=1 + alle IRQs freigeben
		if(lpDIOCParms->portTyp == CANCARD_CAN200)
		{
			wr_can_rr(lpDIOCParms, 8,0x1a);         // Output Control
		}
		else
		{
			wr_can_rr(lpDIOCParms, 8,0xc2);         // Output Control
		}
		wr_can(lpDIOCParms, 0,0x3e);         // Reset Request =0
		wr_can(lpDIOCParms, 10,0x25);		 // Bit ID3 bis ID10
		wr_can(lpDIOCParms, 11,0x88);        // Bit ID0 bis ID2, RTR-Flag,  Data-Length-Code
		
		SetSJA1000Baudrate(lpDIOCParms);
		SetSJA1000AccessCode_Mask(lpDIOCParms);

		if (0x88 == rd_can(lpDIOCParms, 11))
		{
			TRACE(("InitSJA1000 OK"));
			return 1;
		}
	}

	TRACE(("InitSJA1000 nicht OK"));
	return 0;
}

int deInitSJA1000(PCANDEVICE lpDIOCParms)
{
	if (lpDIOCParms->pelican)
	{
		wr_can(lpDIOCParms, 4,0x00);         // alle IRQ aus
	}
	else
	{
		wr_can(lpDIOCParms, 0,0x00);         // Reset Request =0, alle IRQ aus
	}
	return 0;
}

void ConvertBinToText(char *s, long w)
{
	int t;

	for(t=0;t<32;t++)
	{
		if (w & (1<<t)) s[t] = '1';
		else s[t] = '0';
	}
	s[t] = '\n';
	s[t+1] = '\r';
	s[t+2] = 0;
}

unsigned long ConvertHexToInt(const char *string)
{
	unsigned long result = 0;
	int t;
	char z;
	char v;

	for(t=0;string[t];t++)
	{
		z = mytoupper(string[t]);
		
		if ((z>='A') && (z<='F'))  // A...F
		{
			v = z-'A'+10;
		}
		else if ((z>='0') && (z<='9'))  // 0...9
		{
			v = z-'0';
		}
		else	// Error
		{
			return 0;
		}
		result = (result<<4)+v;
	}
	return result;
}

int CompareString(const char *a, const char *b, int case_sensitive)
{
	int t;
	for(t=0;a[t] && b[t];t++)
	{
		if (case_sensitive)
		{
			if (a[t] != b[t]) return FALSE;
		}
		else
		{
			if (mytolower(a[t]) != mytolower(b[t])) return FALSE;
		}
	}
	return TRUE;
}



void CopyMemory(void *d, void *z, unsigned long l)
{
	unsigned char *a,*b;
	unsigned long t;

	a = (unsigned char *)d;
	b = (unsigned char *)z;
	for(t=0;t<l;t++)
		a[t] = b[t];
}

int checkSJA1000(PCANDEVICE lpDIOCParms)
{

	TRACE(("checkSJA1000"));
	if (rd_can(lpDIOCParms, 0) & 0x01)   // Baustein wurde in der Zwischenzeit rckgesetzt
	{
		initSJA1000(lpDIOCParms);  // Setze Baustein auf definierte Werte und benachrichtige Applikation (ActiveX)
						               // die Applikation mu dann die tatschlichen Einstellungen
						               // wieder setze
		/*
		 *   Die Benachrichtigung sollte eigentlich nicht mehr notwendig sein, da die Baudrate und Code und Mask in
		 *   der CANDEVICE-Struktur gespeichert sind
		 *   
		if (lpDIOCParms->admISRInfo && (lpDIOCParms->admISRAktuell >= 0))
		{
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 1;
			Event2UserProgramm(lpDIOCParms, EV2PRG_ERROR);
		}
		*/
		return 0;
	}

	return 1;       // Alles OK
}

void TransmitToSJA1000(PCANDEVICE lpDIOCParms)
{
	unsigned char   i1,i2,len;

	if(lpDIOCParms->standby)
	{
		SJA1000_Commands(lpDIOCParms, CAN_COMMAND_STANDBY_OFF);
	}

	len= (lpDIOCParms->sendedaten.canDLC > 8)?8:lpDIOCParms->sendedaten.canDLC;

	// Achtung! Die Routine wird auch aus der IRQ-Routine heraus aufgerufen
	if (lpDIOCParms->pelican)
	{
		if (lpDIOCParms->eff)
		{
    		TRACE(("TransmitToSJA1000 -1-"));
			i1 = len;
			i1|= lpDIOCParms->sendedaten.canRTR?0x40:0x00;
			i1|= 0x80; // Modus: EFF
			wr_can(lpDIOCParms, 16,i1);
			i1 = (unsigned char)((lpDIOCParms->sendedaten.canIdentifier & 0x1f) << 3);
			wr_can(lpDIOCParms, 20,i1);
			i1 = (unsigned char)((lpDIOCParms->sendedaten.canIdentifier >> 5) & 0xff);
			wr_can(lpDIOCParms, 19,i1);
			i1 = (unsigned char)((lpDIOCParms->sendedaten.canIdentifier >> 13) & 0xff);
			wr_can(lpDIOCParms, 18,i1);
			i1 = (unsigned char)((lpDIOCParms->sendedaten.canIdentifier >> 21) & 0xff);
			wr_can(lpDIOCParms, 17,i1);
			
			for(i1=0;i1<len;i1++)
				wr_can(lpDIOCParms, (unsigned char)21+i1,lpDIOCParms->sendedaten.canMsg[i1]);
			lpDIOCParms->transmitting = 1;
			wr_can(lpDIOCParms, 1,0x01); // Daten senden (Transmission Request)
		}
		else
		{
			// Identifier begrenzen auf 11 Bit (sonst wird spter was falsches angezeigt)
    		TRACE(("TransmitToSJA1000 -2-"));
			lpDIOCParms->sendedaten.canIdentifier &= 0x7ff;  

			i1 = len;
			i1|= lpDIOCParms->sendedaten.canRTR?0x40:0x00;
			wr_can(lpDIOCParms, 16,i1);
			i1 = (unsigned char)((lpDIOCParms->sendedaten.canIdentifier & 7) << 5);
			wr_can(lpDIOCParms, 18,i1);
			i1 = (unsigned char)((lpDIOCParms->sendedaten.canIdentifier >> 3) & 0xff);
			wr_can(lpDIOCParms, 17,i1);
			
			for(i1=0;i1<len;i1++)
				wr_can(lpDIOCParms, (unsigned char)19+i1,lpDIOCParms->sendedaten.canMsg[i1]);
			lpDIOCParms->transmitting = 1;
			wr_can(lpDIOCParms, 1,0x01); // Daten senden (Transmission Request)
		}
	}
	else
	{
		// Identifier begrenzen auf 11 Bit (sonst wird spter was falsches angezeigt)
   		TRACE(("TransmitToSJA1000 -3-"));
		lpDIOCParms->sendedaten.canIdentifier &= 0x7ff;  

		i1 = (unsigned char)((lpDIOCParms->sendedaten.canIdentifier >> 3) & 0xff);
		i2 = (unsigned char)(lpDIOCParms->sendedaten.canIdentifier & 0x07) << 5;
		i2|= lpDIOCParms->sendedaten.canRTR?0x10:0x00;
		i2|= len;
		
		wr_can(lpDIOCParms, 10,i1);
		wr_can(lpDIOCParms, 11,i2);
		
		for(i1=0;i1<len;i1++)
		{
			wr_can(lpDIOCParms, (unsigned char)12+i1,lpDIOCParms->sendedaten.canMsg[i1]);
		}
		lpDIOCParms->transmitting = 1;
		wr_can(lpDIOCParms, 1,0x01); // Daten senden (Transmission Request)
	}
}

void SucheNaechsteISRStruktur(PCANDEVICE lpDIOCParms)
{
	int   t;

	// Suche jetzt nchste freie Info-Struktur
	for(t=1;t<MAX_VISRINFO;t++)
	{
		if (lpDIOCParms->admISRInfo[(lpDIOCParms->admISRAktuell+t)%MAX_VISRINFO].isrNummer == 0) break;
	}
	if (t < MAX_VISRINFO)
		lpDIOCParms->admISRAktuell = (lpDIOCParms->admISRAktuell+t)%MAX_VISRINFO;
	else 
		lpDIOCParms->admISRAktuell = -1; // -1 = keine Struktur ist zur Zeit frei
}

unsigned long IOProc1(PCANDEVICE lpDIOCParms)
{
	CANPARINFO info;

	//TRACE(("IOProc1 %p %d %p %d",lpDIOCParms->lpvOutBuffer,lpDIOCParms->cbOutBuffer,lpDIOCParms->lpvInBuffer,lpDIOCParms->cbInBuffer));
	if (lpDIOCParms->lpvOutBuffer && (lpDIOCParms->cbOutBuffer >= sizeof(info)))
	{
		info.version = CANPAR_VERSION;
		info.time = GetTime(lpDIOCParms);
		info.card = 1;
		info.os   = OPERATING_SYSTEM;
		CopyMemory((void *)lpDIOCParms->lpvOutBuffer,&info,sizeof(info));
		lpDIOCParms->cbBytesReturned = sizeof(info);
	}
	else lpDIOCParms->cbBytesReturned = 0;
    return(NO_ERROR);
}

unsigned long IOProc2(PCANDEVICE lpDIOCParms)
{
    TRACE(("CVXD_W32_Proc2: Schreiben in Register (RR=1)"));
	if (lpDIOCParms->lpvInBuffer)
	{
#ifdef DEMOVERSION
		// Zugriff auf die BTR Register verweigern
		if (0x06 == ((unsigned char*)lpDIOCParms->lpvInBuffer)[0])
		{
			if (lpDIOCParms->admISRInfo && (lpDIOCParms->admISRAktuell >= 0))
			{
				lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
				lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
				Event2UserProgramm(lpDIOCParms, EV2PRG_BAUDRATE_DEMO);
				CreateEvent2UserSpace(lpDIOCParms);
			}
			lpDIOCParms->cbBytesReturned = 0;
			return(ERROR_INVALID);
		}
		if (0x07 == ((unsigned char*)lpDIOCParms->lpvInBuffer)[0])
		{
			if (lpDIOCParms->admISRInfo && (lpDIOCParms->admISRAktuell >= 0))
			{
				lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
				lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
				Event2UserProgramm(lpDIOCParms, EV2PRG_BAUDRATE_DEMO);
				CreateEvent2UserSpace(lpDIOCParms);
			}
			lpDIOCParms->cbBytesReturned = 0;
			return(ERROR_INVALID);
		}
#endif
		wr_can_rr(lpDIOCParms, ((unsigned char*)lpDIOCParms->lpvInBuffer)[0],((unsigned char*)lpDIOCParms->lpvInBuffer)[1]);
	}
	lpDIOCParms->cbBytesReturned = 0;
    return(NO_ERROR);
}

unsigned long IOProc3(PCANDEVICE lpDIOCParms)
{
    TRACE(("CVXD_W32_Proc3: Lesen aus Register (RR=1)"));
	if (lpDIOCParms->lpvInBuffer && lpDIOCParms->lpvOutBuffer)
	{
		((unsigned char*)lpDIOCParms->lpvOutBuffer)[0] = rd_can_rr(lpDIOCParms, ((unsigned char*)lpDIOCParms->lpvInBuffer)[0]);
		lpDIOCParms->cbBytesReturned = 1;
	}
	else lpDIOCParms->cbBytesReturned = 0;
    return(NO_ERROR);
}


unsigned long IOProc4(PCANDEVICE lpDIOCParms)
{
    TRACE(("CVXD_W32_Proc4: Schreiben in Register"));
	if (lpDIOCParms->lpvInBuffer)
	{
#ifdef DEMOVERSION
		// Zugriff auf die BTR Register verweigern
		if (0x06 == ((unsigned char*)lpDIOCParms->lpvInBuffer)[0])
		{
			if (lpDIOCParms->admISRInfo && (lpDIOCParms->admISRAktuell >= 0))
			{
				lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
				lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
				Event2UserProgramm(lpDIOCParms, EV2PRG_BAUDRATE_DEMO);
				CreateEvent2UserSpace(lpDIOCParms);
			}
			lpDIOCParms->cbBytesReturned = 0;
			return(ERROR_INVALID);
		}
		if (0x07 == ((unsigned char*)lpDIOCParms->lpvInBuffer)[0])
		{
			if (lpDIOCParms->admISRInfo && (lpDIOCParms->admISRAktuell >= 0))
			{
				lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
				lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
				Event2UserProgramm(lpDIOCParms, EV2PRG_BAUDRATE_DEMO);
				CreateEvent2UserSpace(lpDIOCParms);
			}
			lpDIOCParms->cbBytesReturned = 0;
			return(ERROR_INVALID);
		}
#endif
		wr_can(lpDIOCParms, ((unsigned char*)lpDIOCParms->lpvInBuffer)[0],((unsigned char*)lpDIOCParms->lpvInBuffer)[1]);
	}
	lpDIOCParms->cbBytesReturned = 0;
    return(NO_ERROR);
}

unsigned long IOProc5(PCANDEVICE lpDIOCParms)
{
    TRACE(("CVXD_W32_Proc5: Lesen aus Register"));
	if (lpDIOCParms->lpvInBuffer && lpDIOCParms->lpvOutBuffer)
	{
		((unsigned char*)lpDIOCParms->lpvOutBuffer)[0] = rd_can(lpDIOCParms, ((unsigned char*)lpDIOCParms->lpvInBuffer)[0]);
		lpDIOCParms->cbBytesReturned = 1;
	}
	else lpDIOCParms->cbBytesReturned = 0;
    return(NO_ERROR);
}

unsigned long IOProc8(PCANDEVICE lpDIOCParms)
{
    TRACE(("CVXD_W32_Proc8: Status auslesen"));
	
	if (lpDIOCParms->lpvOutBuffer && (lpDIOCParms->cbOutBuffer >= 1))
	{
		((unsigned char*)lpDIOCParms->lpvOutBuffer)[0] = rd_can(lpDIOCParms, 2);
		lpDIOCParms->cbBytesReturned = 1;
	}
	else lpDIOCParms->cbBytesReturned = 0;

    return(NO_ERROR);
}

unsigned long IOProc9(PCANDEVICE lpDIOCParms)
{
    unsigned char *tmp;
    unsigned long *tmp2;
	TRACE(("CVXD_W32_Proc9:Acceptance Code + Mask lesen/schreiben"));
	
	checkSJA1000(lpDIOCParms);

	// Falls Werte bergeben wurden -> Schreibe diese Werte in SJA1000-Baustein
	if (lpDIOCParms->lpvInBuffer)
	{
		if(lpDIOCParms->cbInBuffer >= 8)  // Es wurden zwei LONG-Werte bergeben (CODE + MASK) => gltig fr BASIC und PeliCAN
		{
			tmp2 = (unsigned long *)lpDIOCParms->lpvInBuffer;
			lpDIOCParms->access_code = tmp2[0];
			lpDIOCParms->access_mask = tmp2[1];
		}
		else if (lpDIOCParms->cbInBuffer >= 3) // Es wurden zwei BYTE-Werte bergeben (CODE + MASK) => gltig nur fr BASIC
		{
			tmp = (unsigned char *)lpDIOCParms->lpvInBuffer;
			if (tmp[0] & 0x03) // aus Kompatibilittsgrnde
			{
				if (tmp[0] & 0x01)      // Acceptance Code
				{
					lpDIOCParms->access_code = tmp[1];
				}
				if (tmp[0] & 0x02)      // Acceptance Mask
				{
					lpDIOCParms->access_mask = tmp[2];
				}
			}
		}

		SetSJA1000AccessCode_Mask(lpDIOCParms);
	}

	// Kein Ergebnis mehr zurckgeben
	lpDIOCParms->cbBytesReturned = 0;
    return(NO_ERROR);
}


unsigned long IOProc10(PCANDEVICE lpDIOCParms)
{
    struct ISRInfo *tmp;
	short  t,x;

	
	// Ist noch eine Botschaft unterwegs ?
	// Ja -> Botschaft erstaml in den Sendepuffer eintragen
	// NEIN -> Botschaft wird sofort gesendet
	TRACE(("CVXD_W32_Proc10: TransmitData: %d",lpDIOCParms->transmitting));
	if (lpDIOCParms->transmitting)
	{
		TRACE(("CVXD_W32_Proc10: TransmitData -1-"));
		if(lpDIOCParms->sndISRInfo && lpDIOCParms->lpvInBuffer && (lpDIOCParms->cbInBuffer >= sizeof(struct ISRInfo)))
		{
			// Kopiere Struktur in Warteschlange
			tmp = (struct ISRInfo *)lpDIOCParms->lpvInBuffer;
			CopyMemory(&lpDIOCParms->sndISRInfo[lpDIOCParms->sndISRAktuell],tmp,sizeof(struct ISRInfo));

			// Setze aktuellen Zhler um eins hher (Ringpufferstruktur)
			// Vorher wird noch geprft ob berhaupt noch Platz ist (oder ob der
			// Ringpuffer einmal berholt wurde)
			t = (lpDIOCParms->sndISRAktuell+1)%MAX_VISRINFO;
			if (t == lpDIOCParms->sndISRFinished)  // die noch zu bearbeitenden Strukturen eingeholt ?
			{
				return(ERROR_BUFFER_OVERFLOW);
			}

			// Alles OK -> Platz war da fr die nchste Struktur
			// -> Korrigiere jetzt den Zeiger fr den Ringpuffer
			// -> damit es keine Strungen gibt, wird der IRQ kurz maskiert
			x = 1;
			MaskIRQ(lpDIOCParms);
			if (lpDIOCParms->transmitting)                       // Botschaft immer noch unterwegs ?
				lpDIOCParms->sndISRAktuell = t;              // JA -> Botschaft als "Pending" eintragen
			else
				x = 0;                                  // NEIN -> Botschaft wird nicht als "Pending" eingetragen, sondern doch sofort gesendet
			UnmaskIRQ(lpDIOCParms);

			if (!x)
			{
				// Kopiere Daten in die "Sendestruktur"
				CopyMemory(&lpDIOCParms->sendedaten,tmp,sizeof(struct ISRInfo));
				TransmitToSJA1000(lpDIOCParms); // IRQ war bereits beendet -> Sende Botschaft also doch sofort
			}

			return(NO_ERROR);

		}
	}

	checkSJA1000(lpDIOCParms);
	TRACE(("CVXD_W32_Proc10: TransmitData -2-"));
	
	if (lpDIOCParms->lpvInBuffer && (lpDIOCParms->cbInBuffer >= sizeof(struct ISRInfo)))
	{
		tmp = (struct ISRInfo *)lpDIOCParms->lpvInBuffer;
		// Kopiere Daten in die "Sendestruktur"
		CopyMemory(&lpDIOCParms->sendedaten,tmp,sizeof(struct ISRInfo));
		TransmitToSJA1000(lpDIOCParms);
	}
	lpDIOCParms->cbBytesReturned = 0;
    return(NO_ERROR);
}

unsigned long IOProc11(PCANDEVICE lpDIOCParms)
{
	unsigned long retVal=NO_ERROR;
	TRACE(("CVXD_W32_Proc11: Baudrate setzen"));
	checkSJA1000(lpDIOCParms);
	if (lpDIOCParms->lpvInBuffer && (lpDIOCParms->cbInBuffer >= 4))
	{
		lpDIOCParms->baudrate = ((unsigned long *)lpDIOCParms->lpvInBuffer)[0];
		if(!SetSJA1000Baudrate(lpDIOCParms))
			retVal = ERROR_INVALID;
	}
	lpDIOCParms->cbBytesReturned = 0;
    return(retVal);
}

unsigned long IOProc12(PCANDEVICE lpDIOCParms)
{
    TRACE(("CVXD_W32_Proc12: Abort Transmission "));
	lpDIOCParms->sndISRFinished = 0;
	lpDIOCParms->sndISRAktuell = 0;
    wr_can(lpDIOCParms, 1,0x02);
	rd_can(lpDIOCParms, 3);
	lpDIOCParms->transmitting = 0;
	return(NO_ERROR);
}

unsigned long IOProc13(PCANDEVICE lpDIOCParms)
{
	struct ISRInfo *i;
	int tmp;

    TRACE(("CVXD_W32_Proc13: Daten abholen Nr: %ld [%d]",lpDIOCParms->admISRFinished,lpDIOCParms->admISRInfo[lpDIOCParms->admISRFinished].isrNummer));
	if (lpDIOCParms->irq_allowed)
	{
	    //Debug_Printf("FREI admISRaktuell %d %d sndISRAktuell %d %d\n\r",lpDIOCParms->admISRAktuell,lpDIOCParms->admISRFinished,lpDIOCParms->sndISRAktuell,lpDIOCParms->sndISRFinished);
	    //DbgPrint("FREI admISRaktuell %d %d sndISRAktuell %d %d\n",lpDIOCParms->admISRAktuell,lpDIOCParms->admISRFinished,lpDIOCParms->sndISRAktuell,lpDIOCParms->sndISRFinished);
		if (lpDIOCParms->cbOutBuffer >= sizeof(struct ISRInfo))
		{
			i = (struct ISRInfo *)lpDIOCParms->lpvOutBuffer;
			CopyMemory(i, &lpDIOCParms->admISRInfo[lpDIOCParms->admISRFinished], sizeof(struct ISRInfo));
			tmp = lpDIOCParms->admISRFinished;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRFinished].isrNummer = 0;
			if (i->isrNummer)
				lpDIOCParms->admISRFinished = (lpDIOCParms->admISRFinished+1)%MAX_VISRINFO;
			lpDIOCParms->cbBytesReturned = sizeof(struct ISRInfo);

			// War keine Struktur mehr frei ? (zuviele IRQs, die Anwendung hatte keine Zeit fr Reaktion)
			if (lpDIOCParms->admISRAktuell < 0)
			{
				TRACE(("CVXD_W32_Proc13: Speicher freigeben"));
				// Dann setze aktuelle Nummer auf die soeben freigegebene Struktur
				lpDIOCParms->admISRAktuell = tmp;
			}
		}
	}
	return(NO_ERROR);
}

unsigned long IOProc14(PCANDEVICE lpDIOCParms)
{
	unsigned char *ptr;
	TRACE(("CVXD_W32_Proc14: Change to BASIC-CAN/PeliCAN"));
	lpDIOCParms->sndISRFinished = 0;
	lpDIOCParms->sndISRAktuell = 0;

	deInitSJA1000(lpDIOCParms);
	ptr = lpDIOCParms->lpvInBuffer;
	if (ptr && ptr[0])  // PeliCAN
	{
		TRACE(("CVXD_W32_Proc14: Change to PeliCAN"));
		lpDIOCParms->pelican = 1;
		initSJA1000(lpDIOCParms);
	}
	else // BASIC Can
	{
		TRACE(("CVXD_W32_Proc14: Change to Basic-CAN"));
		lpDIOCParms->pelican = 0;
		initSJA1000(lpDIOCParms);
	}
	return(NO_ERROR);
}

unsigned long IOProc15(PCANDEVICE lpDIOCParms)
{
	unsigned char *ptr;
	
	TRACE(("CVXD_W32_Proc15: Change to SFF/EFF"));
	
	lpDIOCParms->sndISRFinished = 0;
	lpDIOCParms->sndISRAktuell = 0;

	ptr = lpDIOCParms->lpvInBuffer;
	if (ptr && ptr[0])  // EFF
	{
		lpDIOCParms->eff = 1;
	}
	else // SFF
	{
		lpDIOCParms->eff = 0;
	}
	return(NO_ERROR);
}

unsigned long IOProc16(PCANDEVICE lpDIOCParms)
{
	unsigned char *ptr;
	
	TRACE(("CVXD_W32_Proc16: Command"));
	
	checkSJA1000(lpDIOCParms);
	ptr = lpDIOCParms->lpvInBuffer;
	if (ptr) 
	{
		return SJA1000_Commands(lpDIOCParms, ptr[0]);
	}
	return ERROR_INVALID;
}

void Event2UserProgramm(PCANDEVICE lpDIOCParms, int msg)
{
	lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].msg = msg;
	lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer++;  // Die Nummer mu im Bereich: 1...n  liegen (0 ist reserviert!)
	SucheNaechsteISRStruktur(lpDIOCParms);
}

int InterruptProcedure(PCANDEVICE lpDIOCParms)
{     
	int ret=0;

	TRACE(("InterruptProcedure"));
	if (lpDIOCParms->pelican)
	{
		while(1)
		{
			if(PeliCanISR(lpDIOCParms))
				ret=1;
			else
				break;
		}
	}
	else
	{
		while(1)
		{
			if(BasicCanISR(lpDIOCParms))
				ret=1;
			else
				break;
		}
	}
	return ret;
}

int PeliCanISR(PCANDEVICE lpDIOCParms)
{                                                       // CSerialPort::HwIntProc
    unsigned char irq;
	unsigned char status;
	unsigned long id;
	int   t;

	//if (intid & 0x40)
	//      return FALSE;                   // this port isn't interrupting

	TRACE(("Pelican IRQ"));
	irq = rd_can(lpDIOCParms, 3);                                                // Interrupt-Register lesen
	if (!(irq & 0xff))
		return FALSE;

	TRACE(("IRQ wird bearbeitet"));
	//Debug_Printf("IRQ admISRaktuell %d %d sndISRAktuell %d %d\n\r",lpDIOCParms->admISRAktuell,lpDIOCParms->admISRFinished,lpDIOCParms->sndISRAktuell,lpDIOCParms->sndISRFinished);
	//DbgPrint("IRQ admISRaktuell %d %d sndISRAktuell %d %d\n",lpDIOCParms->admISRAktuell,lpDIOCParms->admISRFinished,lpDIOCParms->sndISRAktuell,lpDIOCParms->sndISRFinished);
	//TRACE(("IRQ admISRaktuell %d %d sndISRAktuell %d %d\n",lpDIOCParms->admISRAktuell,lpDIOCParms->admISRFinished,lpDIOCParms->sndISRAktuell,lpDIOCParms->sndISRFinished));
	if (lpDIOCParms->admISRInfo && (lpDIOCParms->admISRAktuell >= 0))
	{
		if (irq & 0x01) // Receive Interrupt
		{
			// Flle Info-Struktur
			TRACE(("Receive IRQ Nr. %d",lpDIOCParms->admISRAktuell));
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].status = rd_can(lpDIOCParms, 2);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].lost = lpDIOCParms->admLostReceived;
		    lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].timeStamp = GetTime(lpDIOCParms);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			status = rd_can(lpDIOCParms, 16);
			if (status & 0x80) // Extended Frame Format (EFF)
			{
				id = rd_can(lpDIOCParms, 17);
				id = (id << 8) + rd_can(lpDIOCParms, 18);
				id = (id << 8) + rd_can(lpDIOCParms, 19);
				id = (id << 8) + rd_can(lpDIOCParms, 20);
				id = id >> 3;
				for(t=0;t<8;t++)
				{
					lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canMsg[t] = rd_can(lpDIOCParms, (unsigned char)21+t);
				}
			}
			else	// Standard Frame Format (SFF)
			{
				id = rd_can(lpDIOCParms, 17);
				id = (id << 8) + rd_can(lpDIOCParms, 18);
				id = id >> 5;
				for(t=0;t<8;t++)
				{
					lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canMsg[t] = rd_can(lpDIOCParms, (unsigned char)19+t);
				}
			}
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canIdentifier = id;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canRTR = (status & 0x40) >> 6;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canDLC = (status & 0x0f);
			wr_can(lpDIOCParms, 1,0x04); // ReleaseReceiveBuffer (Rx Buffer wird dadurch ungltig)
			
			Event2UserProgramm(lpDIOCParms, EV2PRG_RECEIVED);
			lpDIOCParms->admLostReceived = 0;
		}
		if (irq & 0x02)                // Transmit Interrupt
		{
			TRACE(("Transmit IRQ"));

			// Kopiere Daten aus Original Sendepuffer
			CopyMemory(&lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell],&lpDIOCParms->sendedaten,sizeof(struct ISRInfo));
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].status = rd_can(lpDIOCParms, 2);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].lost = lpDIOCParms->admLostReceived;
		    lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].timeStamp = GetTime(lpDIOCParms);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			
			Event2UserProgramm(lpDIOCParms, EV2PRG_TRANSMIT);
			
			// Stehen noch Botschaften aus ?
			if (lpDIOCParms->sndISRFinished != lpDIOCParms->sndISRAktuell)
			{
				// Kopiere Daten in die "Sendestruktur"
				CopyMemory(&lpDIOCParms->sendedaten,&lpDIOCParms->sndISRInfo[lpDIOCParms->sndISRFinished],sizeof(struct ISRInfo));
				TransmitToSJA1000(lpDIOCParms);
				lpDIOCParms->sndISRFinished = (lpDIOCParms->sndISRFinished+1)%MAX_VISRINFO;
			}
			else
				lpDIOCParms->transmitting = 0;
		}
		if (irq & 0x04)                   // Error Interrupt
		{
			TRACE(("Error Warning IRQ"));
			lpDIOCParms->transmitting = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].status = rd_can(lpDIOCParms, 2);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].lost = lpDIOCParms->admLostReceived;
		    lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].timeStamp = GetTime(lpDIOCParms);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canIdentifier = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canRTR = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canDLC = 0;
			wr_can(lpDIOCParms, 1,0x0e);                 // Release Receive Buffer + ClearDataOverrun + AbortTransmission (danach erfolgt ein Transmit-IRQ)
			wr_can(lpDIOCParms, 0, ~0x01 & rd_can(lpDIOCParms, 0));       // resetMode lschen
			Event2UserProgramm(lpDIOCParms, EV2PRG_WARNING);
		}
		if (irq & 0x08)                 // Overrun Interrupt
		{
			TRACE(("Overrun IRQ"));
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].status = rd_can(lpDIOCParms, 2);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].lost = lpDIOCParms->admLostReceived;
		    lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].timeStamp = GetTime(lpDIOCParms);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canIdentifier = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canRTR = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canDLC = 0;
			wr_can(lpDIOCParms, 1,0x0c); // Release Receive Buffer + ClearDataOverrun
			Event2UserProgramm(lpDIOCParms, EV2PRG_OVERRUN);
		}
		if (irq & 0x10)                  // WakeUp Interrupt
		{
			TRACE(("WakeUp IRQ"));
			lpDIOCParms->standby=0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].status = rd_can(lpDIOCParms, 2);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].lost = lpDIOCParms->admLostReceived;
		    lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].timeStamp = GetTime(lpDIOCParms);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canIdentifier = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canRTR = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canDLC = 0;
			Event2UserProgramm(lpDIOCParms, EV2PRG_WAKEUP);
		}
		if (irq & 0x20)                   // Error Passive Interrupt
		{
			TRACE(("Error passive IRQ"));
			lpDIOCParms->transmitting = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].status = rd_can(lpDIOCParms, 2);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].lost = lpDIOCParms->admLostReceived;
		    lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].timeStamp = GetTime(lpDIOCParms);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canIdentifier = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canRTR = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canDLC = 0;
			Event2UserProgramm(lpDIOCParms, EV2PRG_PASSIVE);
		}
		if (irq & 0x40)                   // Arbitration lost Interrupt
		{
			TRACE(("Arbitration lost IRQ"));
			lpDIOCParms->transmitting = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].status = rd_can(lpDIOCParms, 2);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].lost = lpDIOCParms->admLostReceived;
		    lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].timeStamp = GetTime(lpDIOCParms);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canIdentifier = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canRTR = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canDLC = 0;
			Event2UserProgramm(lpDIOCParms, EV2PRG_ARBITRATION);
		}
		if (irq & 0x80)                   // Error Passive Interrupt
		{
			TRACE(("Buserror IRQ"));
			lpDIOCParms->transmitting = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].status = rd_can(lpDIOCParms, 2);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].lost = lpDIOCParms->admLostReceived;
		    lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].timeStamp = GetTime(lpDIOCParms);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canIdentifier = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canRTR = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canDLC = 0;
			wr_can(lpDIOCParms, 1,0x0e);                 // Release Receive Buffer + ClearDataOverrun + AbortTransmission (danach erfolgt ein Transmit-IRQ)
			wr_can(lpDIOCParms, 0, ~0x01 & rd_can(lpDIOCParms, 0));       // resetMode lschen
			Event2UserProgramm(lpDIOCParms, EV2PRG_BUSERROR);
		}
	}
	else    // Setze nur die Bits zurck
	{
		if (irq & 0x01) // Receive Interrupt
		{
			wr_can(lpDIOCParms, 1,0x04); // ReleaseReceiveBuffer (Rx Buffer wird dadurch ungltig)
			lpDIOCParms->admLostReceived++;
		}
		if (irq & 0x02)         // Transmit Interrupt
		{
			lpDIOCParms->transmitting = 0;
			
			// Stehen noch Botschaften aus ?
			if (lpDIOCParms->sndISRFinished != lpDIOCParms->sndISRAktuell)
			{
				// Kopiere Daten in die "Sendestruktur"
				CopyMemory(&lpDIOCParms->sendedaten,&lpDIOCParms->sndISRInfo[lpDIOCParms->sndISRFinished],sizeof(struct ISRInfo));
				TransmitToSJA1000(lpDIOCParms);
				lpDIOCParms->sndISRFinished = (lpDIOCParms->sndISRFinished+1)%MAX_VISRINFO;
			}
		}
		if (irq & 0x04)                 // Error Interrupt
		{
			lpDIOCParms->transmitting = 0;
			wr_can(lpDIOCParms, 1,0x0e);                 // Release Receive Buffer + ClearDataOverrun + AbortTransmission (danach erfolgt ein Transmit-IRQ)
			status = rd_can(lpDIOCParms, 0);             // Status einlesen
			status = status & 0xfe; // ResetMode lschen
			wr_can(lpDIOCParms, 0,status);       
		}
		if (irq & 0x08)                 // Overrun Interrupt
		{
			wr_can(lpDIOCParms, 1,0x0c); // Release Receive Buffer + ClearDataOverrun
		}
		if (irq & 0x10)                  // WakeUp Interrupt
		{
			lpDIOCParms->standby=0;
		}
	}
	
	return TRUE;                            // TRUE = IRQ wurde bedient
}

int BasicCanISR(PCANDEVICE lpDIOCParms)
{                                                       // CSerialPort::HwIntProc
    unsigned char irq;
	unsigned char status;
	int   t;

	//if (intid & 0x40)
	//      return FALSE;                   // this port isn't interrupting

	TRACE(("Basic IRQ"));
	irq = rd_can(lpDIOCParms, 3);                                                // Interrupt-Register lesen
	if (!(irq & 0x1f))
		return FALSE;

	TRACE(("IRQ wird bearbeitet"));
	//Debug_Printf("IRQ admISRaktuell %d %d sndISRAktuell %d %d\n\r",lpDIOCParms->admISRAktuell,lpDIOCParms->admISRFinished,lpDIOCParms->sndISRAktuell,lpDIOCParms->sndISRFinished);
	//DbgPrint("IRQ admISRaktuell %d %d sndISRAktuell %d %d\n",lpDIOCParms->admISRAktuell,lpDIOCParms->admISRFinished,lpDIOCParms->sndISRAktuell,lpDIOCParms->sndISRFinished);
	if (lpDIOCParms->admISRInfo && (lpDIOCParms->admISRAktuell >= 0))
	{
		if (irq & 0x01) // Receive Interrupt
		{
			// Flle Info-Struktur
			TRACE(("Receive IRQ"));
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].status = rd_can(lpDIOCParms, 2);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].lost = lpDIOCParms->admLostReceived;
		    lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].timeStamp = GetTime(lpDIOCParms);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canIdentifier = rd_can(lpDIOCParms, 20);
			status = rd_can(lpDIOCParms, 21);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canIdentifier = (lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canIdentifier << 3) + ((status & 0xe0)>>5);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canRTR = (status & 0x10) >> 4;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canDLC = (status & 0x0f);
			for(t=0;t<8;t++)
			{
				lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canMsg[t] = rd_can(lpDIOCParms, (unsigned char)22+t);
			}
			wr_can(lpDIOCParms, 1,0x04); // ReleaseReceiveBuffer (Rx Buffer wird dadurch ungltig)
			
			Event2UserProgramm(lpDIOCParms, EV2PRG_RECEIVED);
			lpDIOCParms->admLostReceived = 0;
		}
		if (irq & 0x02)                // Transmit Interrupt
		{
			TRACE(("Transmit IRQ"));
			CopyMemory(&lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell],&lpDIOCParms->sendedaten,sizeof(struct ISRInfo));
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].status = rd_can(lpDIOCParms, 2);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].lost = lpDIOCParms->admLostReceived;
		    lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].timeStamp = GetTime(lpDIOCParms);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			Event2UserProgramm(lpDIOCParms, EV2PRG_TRANSMIT);
			
			// Stehen noch Botschaften aus ?
			if (lpDIOCParms->sndISRFinished != lpDIOCParms->sndISRAktuell)
			{
				// Kopiere Daten in die "Sendestruktur"
				CopyMemory(&lpDIOCParms->sendedaten,&lpDIOCParms->sndISRInfo[lpDIOCParms->sndISRFinished],sizeof(struct ISRInfo));
				TransmitToSJA1000(lpDIOCParms);
				lpDIOCParms->sndISRFinished = (lpDIOCParms->sndISRFinished+1)%MAX_VISRINFO;
			}
			else
				lpDIOCParms->transmitting = 0;
		}
		if (irq & 0x04)                   // Error Interrupt
		{
			TRACE(("Error IRQ"));
			lpDIOCParms->transmitting = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].status = rd_can(lpDIOCParms, 2);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].lost = lpDIOCParms->admLostReceived;
		    lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].timeStamp = GetTime(lpDIOCParms);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canIdentifier = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canRTR = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canDLC = 0;
			
			wr_can(lpDIOCParms, 1,0x0e);                 // Release Receive Buffer + ClearDataOverrun + AbortTransmission (danach erfolgt ein Transmit-IRQ)
			status = rd_can(lpDIOCParms, 0);             // Status einlesen
			status = status & 0xfe; // ResetMode lschen
			wr_can(lpDIOCParms, 0,status);       
			Event2UserProgramm(lpDIOCParms, EV2PRG_BUSERROR);
		}
		if (irq & 0x08)                 // Overrun Interrupt
		{
			TRACE(("Overrun IRQ"));
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].status = rd_can(lpDIOCParms, 2);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].lost = lpDIOCParms->admLostReceived;
		    lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].timeStamp = GetTime(lpDIOCParms);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canIdentifier = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canRTR = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canDLC = 0;
			wr_can(lpDIOCParms, 1,0x0c); // Release Receive Buffer + ClearDataOverrun
			Event2UserProgramm(lpDIOCParms, EV2PRG_OVERRUN);
		}
		if (irq & 0x10)                  // WakeUp Interrupt
		{
			TRACE(("WakeUp IRQ"));
			lpDIOCParms->standby=0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].intern = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].status = rd_can(lpDIOCParms, 2);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].lost = lpDIOCParms->admLostReceived;
		    lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].timeStamp = GetTime(lpDIOCParms);
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].isrNummer = lpDIOCParms->admISRAktuell;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canIdentifier = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canRTR = 0;
			lpDIOCParms->admISRInfo[lpDIOCParms->admISRAktuell].canDLC = 0;
			Event2UserProgramm(lpDIOCParms, EV2PRG_WAKEUP);
		}
	}
	else    // Setze nur die Bits zurck
	{
		if (irq & 0x01) // Receive Interrupt
		{
			wr_can(lpDIOCParms, 1,0x04); // ReleaseReceiveBuffer (Rx Buffer wird dadurch ungltig)
			lpDIOCParms->admLostReceived++;
		}
		if (irq & 0x02)         // Transmit Interrupt
		{
			lpDIOCParms->transmitting = 0;
			
			// Stehen noch Botschaften aus ?
			if (lpDIOCParms->sndISRFinished != lpDIOCParms->sndISRAktuell)
			{
				// Kopiere Daten in die "Sendestruktur"
				CopyMemory(&lpDIOCParms->sendedaten,&lpDIOCParms->sndISRInfo[lpDIOCParms->sndISRFinished],sizeof(struct ISRInfo));
				TransmitToSJA1000(lpDIOCParms);
				lpDIOCParms->sndISRFinished = (lpDIOCParms->sndISRFinished+1)%MAX_VISRINFO;
			}
		}
		if (irq & 0x04)                 // Error Interrupt
		{
			lpDIOCParms->transmitting = 0;
			wr_can(lpDIOCParms, 1,0x0e);                 // Release Receive Buffer + ClearDataOverrun + AbortTransmission (danach erfolgt ein Transmit-IRQ)
			status = rd_can(lpDIOCParms, 0);             // Status einlesen
			status = status & 0xfe; // ResetMode lschen
			wr_can(lpDIOCParms, 0,status);       
		}
		if (irq & 0x08)                 // Overrun Interrupt
		{
			wr_can(lpDIOCParms, 1,0x0c); // Release Receive Buffer + ClearDataOverrun
		}
		if (irq & 0x10)                  // WakeUp Interrupt
		{
			lpDIOCParms->standby=0;
		}
	}
	
	return TRUE;                            // TRUE = IRQ wurde bedient
}

unsigned long Can_OpenDevice(PCANDEVICE lpDIOCParms)
{
	TRACE(("LPTx wird initialisiert"));

	lpDIOCParms->admISRInfo=NULL;
	lpDIOCParms->admISRAktuell=0;
	lpDIOCParms->admLostReceived=0;
	lpDIOCParms->admISRFinished=0;
	
	lpDIOCParms->sndISRInfo=NULL;
	lpDIOCParms->sndISRAktuell=0;
	lpDIOCParms->sndISRFinished=0;
	lpDIOCParms->standby=0;

	TRACE(("OpenDevice - Settings: Baudrate %ld Code %08lx Mask %08lx Pelican %d  ExtendedFrame %d",lpDIOCParms->baudrate,lpDIOCParms->access_code,lpDIOCParms->access_mask,lpDIOCParms->pelican,lpDIOCParms->eff));
	TRACE(("CVXD_Dynamic_Init Memory anfordern (Eventpuffer)"));
	
	lpDIOCParms->admISRInfo = (struct ISRInfo *)Allocate_Memory(MAX_VISRINFO*sizeof(struct ISRInfo));
	if (!lpDIOCParms->admISRInfo) goto error;

	TRACE(("CVXD_Dynamic_Init Memory anfordern (Sendepuffer)"));
	lpDIOCParms->sndISRInfo = (struct ISRInfo *)Allocate_Memory(MAX_VISRINFO*sizeof(struct ISRInfo));
	if (!lpDIOCParms->sndISRInfo) goto error;

	initParallelPort(lpDIOCParms);
	initSJA1000(lpDIOCParms);
	lpDIOCParms->transmitting = 0;       

	TRACE(("Ergebnis: Alles OK"));
	return 0;

error:
	if (lpDIOCParms->admISRInfo)
	{
		TRACE(("Speicher freigeben"));
		Free_Memory(lpDIOCParms->admISRInfo);
		lpDIOCParms->admISRInfo = NULL;
	}
	lpDIOCParms->admISRAktuell = 0;

	if (lpDIOCParms->sndISRInfo)
	{
		TRACE(("SENDE-Speicher freigeben"));
		Free_Memory(lpDIOCParms->sndISRInfo);
		lpDIOCParms->sndISRInfo = NULL;
	}
	lpDIOCParms->sndISRAktuell = 0;


	TRACE(("Rckkehr mit Fehlerkode"));
	return -1;
}

unsigned long Can_CloseDevice(PCANDEVICE lpDIOCParms)
{
    TRACE(("CVXD_CloseDevice"));
	
	deInitSJA1000(lpDIOCParms);
	deinitParallelPort(lpDIOCParms);

	if (lpDIOCParms->admISRInfo)
	{
		TRACE(("Speicher freigeben"));
		Free_Memory(lpDIOCParms->admISRInfo);
		lpDIOCParms->admISRInfo = NULL;
	}
	lpDIOCParms->admISRAktuell = 0;

	if (lpDIOCParms->sndISRInfo)
	{
		TRACE(("SENDE-Speicher freigeben"));
		Free_Memory(lpDIOCParms->sndISRInfo);
		lpDIOCParms->sndISRInfo = NULL;
	}
	lpDIOCParms->sndISRAktuell = 0;

    return 0;
}


