#define EXPORT_SYMTAB
#include <linux/module.h>
#include <linux/malloc.h>
#include <linux/parport.h>
#include <linux/ioctl.h>
#include <asm/uaccess.h>
#include <linux/spinlock.h>
#include <linux/delay.h>
#include <linux/time.h>

#include "../canpar.h"
#include "../utreiber.h"
#include "treiber.h"

/* 		Installationsanleitung LINUX:
 * 		
 * 		char-major Nummern ab 240 sind fr "LOCAL/EXPERIMENTAL" (siehe /usr/src/linux/Documentation/devices.txt)
 * 		=> die Nummer 240 war bei mir bereits belegt!
 * 		in /etc/modules.conf eintragen:  
 * 		--------------------------------------------------
 * 		alias char-major-241     candrv
 * 		pre-install candrv modprobe "-k" parport_pc
 * 		options candrv canpar_major=241
 * 		post-remove candrv rmmod parport_pc parport
 * 	    --------------------------------------------------
 * 	    
 * 	    Nach einer nderung von modules.conf mu folgendes aufgerufen werden:
 * 	    depmod -a
 * 	    
 * 	    
 * 	    Danach ein Device unter /dev anlegen:
 * 		mknod /dev/canpar c 241 0
 * 		
 * 		Damit auch "Otto-Normaluser" Zugriff hat, noch die Zugriffsrechte anpassen:
 * 		(hier hat danach jeder Zugriff!)
 * 		chmod 777 /dev/canpar 
 * 
*/

int canpar_major=0;
unsigned long canpar_baudrate=20000;
int canpar_cardtype=0; // 0=elektor, 1=can200
unsigned long canpar_access_code=0xffffffff;
unsigned long canpar_access_mask=0xffffffff;
unsigned char canpar_20b_mode=0; // 0= BASIC Can 1=Pelican
unsigned char canpar_extended_frame_fmt=0; // 0=11Bit 1=29Bit
CANDEVICE candevice[PARPORT_MAX];
static int canpar_fasync(int fd, struct file *file, int mode);

void wr_can(PCANDEVICE lpDIOCParms, unsigned char adr,unsigned char wert) // Wert in Register des 82C200 an Adresse adr schreiben
{
	//TRACE(("wr_can %u %u",adr,wert));
	MaskIRQ(lpDIOCParms);
	wr_data(lpDIOCParms, adr,wert);
	UnmaskIRQ(lpDIOCParms);
}

void wr_can_rr(PCANDEVICE lpDIOCParms, unsigned char adr,unsigned char wert) // Wert in Register des 82C200 an Adresse adr schreiben
{
	//TRACE(("wr_can %u %u",adr,wert));
	MaskIRQ(lpDIOCParms);
	SetToResetMode(lpDIOCParms);
	wr_data(lpDIOCParms, adr,wert);
	SetToNormalMode(lpDIOCParms);
	UnmaskIRQ(lpDIOCParms);
}

unsigned char rd_can(PCANDEVICE lpDIOCParms, unsigned char adr) // Wert aus Register des 82C200 auslesen
{
	unsigned char bw;
	MaskIRQ(lpDIOCParms);
    bw = rd_data(lpDIOCParms, adr);
	UnmaskIRQ(lpDIOCParms);
	return(bw);
}

unsigned char rd_can_rr(PCANDEVICE lpDIOCParms, unsigned char adr) // Wert aus Register des 82C200 auslesen
{
	unsigned char bw;
	MaskIRQ(lpDIOCParms);
	SetToResetMode(lpDIOCParms);
    bw = rd_data(lpDIOCParms, adr);
	SetToNormalMode(lpDIOCParms);
	UnmaskIRQ(lpDIOCParms);
	return(bw);
}

HIGHLOW GetTime(PCANDEVICE lpDIOCParms)
{
	HIGHLOW ret;
	struct timeval tv;
	do_gettimeofday(&tv);
	ret.low = tv.tv_usec;
	ret.high = tv.tv_sec;
	return ret;
}

void *Allocate_Memory(unsigned long s)
{
	unsigned char *x;
	unsigned long t;
	x = kmalloc(s,GFP_KERNEL);
	if (x)
	{
		for(t=0;t<s;t++)
			x[t] = 0;
	}
	return x;
}

void Free_Memory(void *p)
{
	kfree(p);
}

void MaskIRQ(PCANDEVICE lpDIOCParms)
{
	spin_lock_irqsave(&lpDIOCParms->sja_lock,lpDIOCParms->lock_flags);
	
}

void UnmaskIRQ(PCANDEVICE lpDIOCParms)
{
	spin_unlock_irqrestore(&lpDIOCParms->sja_lock,lpDIOCParms->lock_flags);
}

void initParallelPort(PCANDEVICE lpDIOCParms)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	out2Controlport(lpDIOCParms,0x04); // Schnittstelle initialisieren
	parport_enable_irq(lpDIOCParms->can_device->port);
#else
	if (lpDIOCParms->ecp == 1)  // ECP Modus umschalten
		out2ECPControlport(lpDIOCParms,inECPControlport(lpDIOCParms)|0x20);

	out2Controlport(lpDIOCParms,0x14); // Schnittstelle initialisieren/IRQ ein
#endif
	lpDIOCParms->irq_allowed=1;
/* 
 diesen Teil in die Windows-Treiber kopieren
	if (lpDIOCParms->ecp == 1)
		out2ECPControlport(lpDIOCParms,inECPControlport(lpDIOCParms)|0x20);
	initParallelPort(lpDIOCParms);
	//out2Controlport(lpDIOCParms,0x14); // Schnittstelle initialisieren/IRQ ein
*/
}

void deinitParallelPort(PCANDEVICE lpDIOCParms)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	out2Controlport(lpDIOCParms,inControlport(lpDIOCParms)&0xef); // Schnittstelle: IRQ aus
	parport_disable_irq(lpDIOCParms->can_device->port);
#else
	out2Controlport(lpDIOCParms,inControlport(lpDIOCParms)&0xef); // Schnittstelle: IRQ aus
#endif
	lpDIOCParms->irq_allowed=0;
/*
 diesen Teil in die Windows-Treiber kopieren
	out2Controlport(lpDIOCParms,inControlport(lpDIOCParms)&0xef); // Schnittstelle: IRQ aus
*/
}

void out2Statusport(PCANDEVICE lpDIOCParms, int data)
{                                               
	// Statusregister ist ReadOnly
}                                                       

unsigned char inStatusport(PCANDEVICE lpDIOCParms)
{                                               
	unsigned char ret;
    ret = parport_read_status(lpDIOCParms->can_device->port);
	return ret;
}                                                       

void out2Controlport(PCANDEVICE lpDIOCParms, int data)
{                                               
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    parport_write_control(lpDIOCParms->can_device->port,data & ~0x20);
   	if(data & 0x20)
    		parport_data_reverse(lpDIOCParms->can_device->port);
   	else
    		parport_data_forward(lpDIOCParms->can_device->port);
#else
    parport_write_control(lpDIOCParms->can_device->port,data);
#endif
	lpDIOCParms->statusport = data;

}                                                       

unsigned char inControlport(PCANDEVICE lpDIOCParms)
{                      
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	return lpDIOCParms->statusport;
#else
	unsigned char ret;
    ret = parport_read_control(lpDIOCParms->can_device->port);
	return ret;
#endif
}                                                       

void out2Data(PCANDEVICE lpDIOCParms, int data)
{                                               
    parport_write_data(lpDIOCParms->can_device->port,data);
}                                                       

unsigned char inData(PCANDEVICE lpDIOCParms)
{                                               
	unsigned char ret;
    ret = parport_read_data(lpDIOCParms->can_device->port);
	return ret;
}                                                       

void out2ECPControlport(PCANDEVICE lpDIOCParms, int data)
{                                               
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
    parport_write_econtrol(lpDIOCParms->can_device->port,data);
#endif
}                                                       

unsigned char inECPControlport(PCANDEVICE lpDIOCParms)
{                                               
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	unsigned char ret;
    ret = parport_read_econtrol(lpDIOCParms->can_device->port);
	return ret;
#else
	return 0;
#endif
}     
                                                  
//*************************************************************
void can_irq(int irq, void *dev_id, struct pt_regs *regs)
{
		unsigned int minor=(unsigned int)dev_id;

		// Folgende Abfragen nur whrend der Testphase
		if(minor < 0)
			return;
		if(minor >= PARPORT_MAX)
			return;
			
		candevice[minor].in_isr = TRUE;
		TRACE(("InterruptProcedure -1-"));
    	if (!InterruptProcedure(&candevice[minor]))
		{
			TRACE(("InterruptProcedure -2-"));
			candevice[minor].in_isr = FALSE;
			return;
		}
		TRACE(("InterruptProcedure -3-"));
		candevice[minor].in_isr = FALSE;

		if(candevice[minor].async_queue)
		{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
			kill_fasync(&candevice[minor].async_queue,SIGIO,POLL_IN); /* 2.4.0 */
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,18)
			kill_fasync(candevice[minor].async_queue,SIGIO,POLL_IN); /* 2.2.18 */
#else
			kill_fasync(candevice[minor].async_queue,SIGIO);   /* 2.2.16 */
#endif
#endif
		}
}

int can_preempt(void *handle)
{
  return 1; // 1=nicht freigeben
}

static int canpar_open(struct inode * inode, struct file * file)
{
	unsigned int minor = MINOR(inode->i_rdev);

	if(candevice[minor].open)
		return -ENODEV;
		
	file->private_data = (void *)minor; // wird in "fasync()" bentigt
	
	MOD_INC_USE_COUNT;
	if(parport_claim(candevice[minor].can_device))
	{
		printk(KERN_INFO "CANPAR: No access to PARPORT device\n");
		return -ENODEV;
	}
	
	candevice[minor].max_inOutBuffer = 200;
	candevice[minor].lpvOutBuffer = Allocate_Memory(candevice[minor].max_inOutBuffer);
	if(!candevice[minor].lpvOutBuffer)
		goto error1;
	candevice[minor].lpvInBuffer= Allocate_Memory(candevice[minor].max_inOutBuffer);
	if(!candevice[minor].lpvInBuffer)
		goto error2;

	candevice[minor].ecp = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	if(candevice[minor].can_device->port->modes & (PARPORT_MODE_PCECP+PARPORT_MODE_PCECPEPP+PARPORT_MODE_PCECPPS2))
	{
		printk(KERN_INFO "CANPAR: Using ECP mode\n");
		candevice[minor].ecp = 1;
	}
	else
	{
		printk(KERN_INFO "CANPAR: Using Standard mode (Bidirectional, EPP)\n");
		candevice[minor].ecp = 0;
	}
#endif
	Can_OpenDevice(&candevice[minor]);
	TRACE(("Nach can_openDevice() %p %p",candevice[minor].lpvOutBuffer,candevice[minor].lpvInBuffer));
	candevice[minor].open = 1;
	return 0;
error2:
	printk(KERN_INFO "CANPAR: Out of memory while allocation Input buffer\n");
	Free_Memory(candevice[minor].lpvInBuffer);
	candevice[minor].lpvInBuffer = NULL;
error1:
	printk(KERN_INFO "CANPAR: Out of memory while allocation Output buffer\n");
	Free_Memory(candevice[minor].lpvOutBuffer);
	candevice[minor].lpvOutBuffer = NULL;
	
	return -ENOMEM;
}

static int canpar_release(struct inode * inode, struct file * file)
{
	unsigned int minor = MINOR(inode->i_rdev);
	
	Can_CloseDevice(&candevice[minor]);
	if(candevice[minor].lpvOutBuffer) Free_Memory(candevice[minor].lpvOutBuffer);
	candevice[minor].lpvOutBuffer = 0;
	if(candevice[minor].lpvInBuffer) Free_Memory(candevice[minor].lpvInBuffer);
	candevice[minor].lpvInBuffer = 0;
	parport_release(candevice[minor].can_device);
	canpar_fasync(-1,file,0);
	candevice[minor].open = 0;
	MOD_DEC_USE_COUNT;
	return 0;
}

static void CopyDataFromUserSpace(PCANDEVICE dev, struct ioctlTransfer *iot)
{
	if(!iot)
		return;
		
	if (iot->cbInBuffer && iot->lpvInBuffer)
	{
		copy_from_user(dev->lpvInBuffer,iot->lpvInBuffer,iot->cbInBuffer);
		dev->cbInBuffer= iot->cbInBuffer;
	}
	dev->cbOutBuffer= iot->cbOutBuffer;
	//TRACE(("CopyFromUser %p %d %p %d",dev->lpvOutBuffer,dev->cbOutBuffer,dev->lpvInBuffer,dev->cbInBuffer));
}

static void CopyDataToUserSpace(PCANDEVICE dev, struct ioctlTransfer *iot)
{
	if(!iot)
		return;

	if (iot->cbOutBuffer && dev->cbBytesReturned && iot->lpvOutBuffer)
	{
		copy_to_user(iot->lpvOutBuffer,dev->lpvOutBuffer,iot->cbOutBuffer);
		iot->cbOutBuffer= dev->cbOutBuffer;
		iot->cbBytesReturned = dev->cbBytesReturned;
	}
}

static int canpar_ioctl(struct inode *inode, struct file *file,
		    unsigned int cmd, unsigned long arg)
{
	int status;
	unsigned int minor = MINOR(inode->i_rdev);
	struct ioctlTransfer *iot= (struct ioctlTransfer *)arg;
	struct ioctlTransfer dummyTrans;


	TRACE(("lp%u ioctl, cmd: 0x%x, arg: 0x%lx", minor, cmd, arg));
	candevice[minor].cbBytesReturned= 0;
	candevice[minor].cbOutBuffer	= 0;
	candevice[minor].cbInBuffer		= 0;

	if(!iot) 
	{
		dummyTrans.lpvInBuffer 		= 0;
		dummyTrans.lpvOutBuffer 	= 0;
		dummyTrans.cbInBuffer 		= 0;
		dummyTrans.cbOutBuffer 		= 0;
		iot = &dummyTrans;
	}
	else
	{
		status = verify_area(VERIFY_WRITE,iot,sizeof(struct ioctlTransfer));
		if(status < 0) 
		{
			TRACE(("Fehler nach verify_area -1- : %d",status));
			return status;
		}

		status = verify_area(VERIFY_READ,iot->lpvInBuffer,iot->cbInBuffer);
		if(status < 0) 
		{
			TRACE(("Fehler nach verify_area -2- : %d",status));
			return status;
		}

		status = verify_area(VERIFY_WRITE,iot->lpvOutBuffer,iot->cbOutBuffer);
		if(status < 0) 
		{
			TRACE(("Fehler nach verify_area -3- : %d",status));
			return status;
		}
	}	

	status = 0;
	switch(cmd)
	{
	case IOCTL_IO_1:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc1(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	case IOCTL_IO_2:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc2(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	case IOCTL_IO_3:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc3(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	case IOCTL_IO_4:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc4(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	case IOCTL_IO_5:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc5(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	case IOCTL_IO_7:
		status = 0;
		break;
	case IOCTL_IO_8:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc8(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	case IOCTL_IO_9:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc9(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	case IOCTL_IO_10:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc10(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	case IOCTL_IO_11:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc11(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	case IOCTL_IO_12:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc12(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	case IOCTL_IO_13:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc13(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	case IOCTL_IO_14:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc14(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	case IOCTL_IO_15:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc15(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	case IOCTL_IO_16:
	 	CopyDataFromUserSpace(&candevice[minor], iot);
		status = IOProc16(&candevice[minor]);
	 	CopyDataToUserSpace(&candevice[minor], iot);
		break;
	default:
		TRACE(("Unbekannter IOControlCode aufgerufen"));
		status = -EINVAL;
	}
	TRACE(("ParIIOControl -4- (%d zurck)",candevice[minor].cbBytesReturned));
	return status;
}

static int canpar_fasync(int fd, struct file *file, int mode)
{
	unsigned int minor = (unsigned int)file->private_data;
	// Folgende Abfragen nur whrend der Testphase
	if(minor < 0)
		return 0;
	if(minor >= PARPORT_MAX)
		return 0;
			
	return fasync_helper(fd,file,mode,&candevice[minor].async_queue);
}

static struct file_operations canpar_fops = {         /* siehe in linux/fs.h */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	THIS_MODULE,
#endif
	NULL, //canpar_lseek,
	NULL, //canpar_read,
	NULL, //canpar_write,
	NULL,		/* lp_readdir */
	NULL,		/* lp_poll */
	canpar_ioctl,
	NULL,		/* lp_mmap */
	canpar_open,
	NULL,		/* flush */
	canpar_release, /* release */
	NULL,	/* fsync */
	canpar_fasync, /* fasync */
};

int init_module(void)
{
	int result;
	int num;
	struct parport *port;

	printk("CANPAR: init_module()\n");
    result = register_chrdev(canpar_major,"canpar",&canpar_fops);
    if (result < 0)
    {
    	printk("CANPAR: Can't get MAJOR Number: %d\n",canpar_major);
    	return result;
    }
    
  	for(num=0,port=parport_enumerate(); port && (num < PARPORT_MAX); port = port->next,num++)
  	{
  		candevice[num].open = 0;
		candevice[num].async_queue = 0;
    	candevice[num].can_device = parport_register_device(port, "canpar", can_preempt, NULL, can_irq,0,(void *)num);
    	candevice[num].pelican = canpar_20b_mode;
        candevice[num].portTyp = canpar_cardtype;
    	candevice[num].eff = canpar_extended_frame_fmt;
    	candevice[num].baudrate = canpar_baudrate;
    	candevice[num].access_code = canpar_access_code;
    	candevice[num].access_mask = canpar_access_mask;
    	if (!candevice[num].can_device)
    	{
      		printk("CANPAR: Can't register device no.: %d\n",num);
    	}
    	candevice[num].sja_lock = SPIN_LOCK_UNLOCKED;
  	}
  	return 0;
}

void cleanup_module(void)
{
    int num;
	struct parport *port;
    	
  	printk("CANPAR: cleanup\n");
  	for(num=0,port=parport_enumerate(); port && (num < PARPORT_MAX); port = port->next,num++)
  	{
  		if (candevice[num].can_device)
  			parport_unregister_device(candevice[num].can_device);
  		candevice[num].can_device = 0;
  	}
	unregister_chrdev(canpar_major,"canpar");
}

MODULE_PARM(canpar_20b_mode,"1i");
MODULE_PARM(canpar_extended_frame_fmt,"1i");
MODULE_PARM(canpar_baudrate,"1-4i");
MODULE_PARM(canpar_access_code,"1-4i");
MODULE_PARM(canpar_access_mask,"1-4i");
MODULE_PARM(canpar_cardtype,"1i");
MODULE_PARM(canpar_major,"1-4i");
MODULE_AUTHOR("Jrgen Eder");
MODULE_DESCRIPTION("CAN Bus driver for Elektor CAN Card 6/2000");

