#include <stdio.h>
#include "mcc.h"

//------------------------------------------------------------------------------------------
//
//                       B U T T O N
//
//------------------------------------------------------------------------------------------
void CButton::CreateButton(PCTSTR title)
{
  m_pWindow = gtk_button_new_with_label(title);
  m_pLabel.m_pWindow = GTK_BIN(m_pWindow)->child;
}

void CButton::Connect(void)
{
  //TRACE("CButton::Connect() %p %p\n",this,m_pWindow);
  gtk_signal_connect(GTK_OBJECT(m_pWindow),"clicked",GTK_SIGNAL_FUNC(CS_ButtonClick),this);
  gtk_signal_connect(GTK_OBJECT(m_pWindow),"pressed",GTK_SIGNAL_FUNC(CS_ButtonPressed),this);
  gtk_signal_connect(GTK_OBJECT(m_pWindow),"released",GTK_SIGNAL_FUNC(CS_ButtonReleased),this);
  CGenericButton::Connect();
}

void CButton::Click(void)
{
  //TRACE("CButton::Click %p %p\n",this,m_pParent);
  if (m_pParent)
  {
	  //TRACE("Rufe Parent->Command auf\n");
      m_pParent->Command(m_Id);
  }
}

void CButton::Press(void)
{
  //TRACE("CButton::Press\n");
}

void CButton::Release(void)
{
  //TRACE("CButton::Release\n");
}

void CButton::SetText(PCTSTR text)
{
  m_pLabel.SetText(text);
}
//------------------------------------------------------------------------------------------
//
//                       T O G G L E B U T T O N
//
//------------------------------------------------------------------------------------------
void CToggleButton::CreateButton(PCTSTR title)
{
  m_pWindow = gtk_toggle_button_new_with_label(title);
}

void CToggleButton::Connect(void)
{
  gtk_signal_connect(GTK_OBJECT(m_pWindow),"toggled",GTK_SIGNAL_FUNC(CS_ButtonToggled),this);
  CButton::Connect();
}

void CToggleButton::Toggle(void)
{
  //TRACE("CToggleButton::Toggle\n");
}



//------------------------------------------------------------------------------------------
//
//                       C H E C K B U T T O N
//
//------------------------------------------------------------------------------------------

void CCheckButton::CreateButton(PCTSTR title)
{
  m_pWindow = gtk_check_button_new_with_label(title);
}


void CCheckButton::SetCheck(int markiert)
{
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(m_pWindow),markiert?TRUE:FALSE);
}

int CCheckButton::GetCheck(void)
{
	return gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(m_pWindow))?TRUE:FALSE;
}

//------------------------------------------------------------------------------------------
//
//                       L A B E L
//
//------------------------------------------------------------------------------------------
void CLabel::CreateButton(PCTSTR title)
{
  m_pWindow = gtk_label_new(title);
}

void CLabel::SetText(PCTSTR text)
{
  gtk_label_set(GTK_LABEL(m_pWindow),text);
}

void CLabel::Connect(void)
{
  // nichts tun
}



//------------------------------------------------------------------------------------------
//
//                       C L I S T
//
//------------------------------------------------------------------------------------------
void CList::Init(void)
{
	//CGenericButton::Init();
   m_nTitles = 0;
   m_pTitles = NULL;
   m_nRows   = 0;
}

CList::CList()
{
   Init();	
}

int CList::Create(CScrolledWindow *pParent, int Id, gint nTitles, gchar *texte[], BOOL show)
{
  m_nTitles = nTitles;
  m_pTitles = texte;
  m_pParent = pParent;
  m_Id = Id;
  CreateButton("CLIST"); // Pseude-Name
  if (!m_pWindow) return -1;
  if (m_pParent) gtk_container_add(GTK_CONTAINER(m_pParent->m_pWindow),m_pWindow);
  Connect();
  Show(show);    
  return 0;
}

void CList::CreateButton(PCTSTR title)
{
  if (m_pTitles)
  	m_pWindow = gtk_clist_new_with_titles(m_nTitles,m_pTitles);
  else
  	m_pWindow = gtk_clist_new(m_nTitles);
}

void CList::Clear(void)
{
   m_nRows = 0;
   if(!m_pWindow) return;
   gtk_clist_clear(GTK_CLIST(m_pWindow));
}

void CList::Connect(void)
{
}

int CList::Append(gchar *texte[])
{
   m_nRows++;
   if(!m_pWindow) return 0;
   return gtk_clist_append(GTK_CLIST(m_pWindow),texte);
}

int CList::Prepend(gchar *texte[])
{
   m_nRows++;
   if(!m_pWindow) return 0;
   return gtk_clist_prepend(GTK_CLIST(m_pWindow),texte);
}

int CList::Insert(gint row, gchar *texte[])
{
  m_nRows++;
   if(!m_pWindow) return 0;
   return gtk_clist_insert(GTK_CLIST(m_pWindow),row,texte);
}

void CList::Remove(gint row)
{
   m_nRows--;
   if(!m_pWindow) return;
   gtk_clist_remove(GTK_CLIST(m_pWindow),row);
}

void CList::ShowTitle(BOOL show)
{
   if(!m_pWindow) return;

   if (show)
   		gtk_clist_column_titles_show(GTK_CLIST(m_pWindow));
   else
   		gtk_clist_column_titles_hide(GTK_CLIST(m_pWindow));
}

void CList::SetColumnActive(gint col, BOOL active)
{
   if(!m_pWindow) return;

   if (active)
   		gtk_clist_column_title_active(GTK_CLIST(m_pWindow),col);
   else
   		gtk_clist_column_title_passive(GTK_CLIST(m_pWindow),col);
}

void CList::SetActive(BOOL active)
{
   int t;
   for(t=0;t<m_nTitles;t++)
   		SetColumnActive(t,active);
}

gint CList::GetOptimalColumnWidth(gint column)
{
   if(!m_pWindow) return 0;

   return gtk_clist_optimal_column_width(GTK_CLIST(m_pWindow),column);
}

void CList::SetColumnWidth(gint column, gint width)
{
   if(!m_pWindow) return;

   gtk_clist_set_column_width(GTK_CLIST(m_pWindow),column,width);
}

void CList::Click(void)
{
}

int CList::GetRows(void)
{
   return m_nRows;
}




//------------------------------------------------------------------------------------------
//
//                       L I S T B O X
//
//------------------------------------------------------------------------------------------
void CListBox::CreateButton(PCTSTR title)
{
  m_pWindow = gtk_list_new();
}

void CListBox::AddItem(PCTSTR title)
{
  GtkWidget *item;
  item = gtk_list_item_new_with_label(title);
  if (!item) return;
  gtk_container_add(GTK_CONTAINER(m_pWindow),item);
  gtk_widget_show(item);
}

void CListBox::Clear(int nStart, int nEnd)
{
  gtk_list_clear_items(GTK_LIST(m_pWindow),nStart,nEnd);
}

void CListBox::Connect(void)
{
  gtk_signal_connect(GTK_OBJECT(m_pWindow),"selection_changed",GTK_SIGNAL_FUNC(CS_ButtonClick),this);
  //CGenericButton::Connect();
}

void CListBox::Click(void)
{
  if (m_pParent)
  {
      m_pParent->Command(m_Id);
  }
}

//------------------------------------------------------------------------------------------
//
//                       C O M B O B O X
//
//------------------------------------------------------------------------------------------
CComboBox::CComboBox()
{
	m_Liste=NULL;
	m_Num = -1;
}

CComboBox::~CComboBox()
{
	//TRACE("CComboBox Destruktor\n");
   	Clear();
}

void CComboBox::Destroy(void)
{
   Clear();
   if (m_pWindow)
   {
   		gtk_signal_disconnect_by_data(GTK_OBJECT(GTK_COMBO(m_pWindow)->entry),this);
   		gtk_signal_disconnect_by_data(GTK_OBJECT(GTK_COMBO(m_pWindow)->list),this);
   }
   CGenericButton::Destroy();
}

void CComboBox::CreateButton(PCTSTR title)
{
  m_pWindow = gtk_combo_new();
}

void CComboBox::AddItem(PCTSTR title)
{
  GList *l;
  l = g_list_append(m_Liste,title);
  if(!m_Liste) m_Liste = l;
  gtk_combo_set_popdown_strings(GTK_COMBO(m_pWindow),m_Liste);
}

void CComboBox::Clear(int nStart, int nEnd)
{
  if (m_pWindow)
  	gtk_list_clear_items(GTK_LIST(GTK_COMBO(m_pWindow)->list),nStart,nEnd);
}

void CComboBox::Clear(void)
{
  if (m_Liste)
  {
  	g_list_free(m_Liste);
  	m_Liste = 0;
  	//m_Num=-1;
  }
}

void CComboBox::Connect(void)
{
  GtkList *list;
  list = GTK_LIST(GTK_COMBO(m_pWindow)->list);
  gtk_signal_connect(GTK_OBJECT(GTK_COMBO(m_pWindow)->entry),"changed",GTK_SIGNAL_FUNC(CS_EntryChanged),this);
  gtk_signal_connect(GTK_OBJECT(GTK_COMBO(m_pWindow)->list),"selection_changed",GTK_SIGNAL_FUNC(CS_SelectionChanged),this);
  gtk_signal_connect(GTK_OBJECT(GTK_COMBO(m_pWindow)->list),"button-press-event",GTK_SIGNAL_FUNC(CS_ButtonPressed2),this);
  gtk_signal_connect(GTK_OBJECT(GTK_COMBO(m_pWindow)->list),"button-release-event",GTK_SIGNAL_FUNC(CS_ButtonReleased2),this);
  //CGenericButton::Connect();
}

void CComboBox::Release(void)
{
	/* Wenn eine Liste vorhanden ist -> jetzt abklappern und nachsehen
	 * welcher Eintrag ausgewhlt wurde.
	 * Leider erhlt man keine "ausgewhlte-Item-Nummer" der ComboBox
	 */
  m_Num = -1;
  if(m_Liste)
  {
	GList *list;
	char *text;
	int t=0;
	text = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(m_pWindow)->entry));
	if (text && text[0])
	{
		list = g_list_first(m_Liste);
		while(list)
		{
			if ((strlen(text) == strlen((PCTSTR)list->data)) && !strcmp(text,(PCTSTR)list->data))
			{
  				m_Num = t;
  				//TRACE("CComboBox::Click %d %s\n",m_Num,text);
				break;
			}
			t++;
			list = g_list_next(list);
		}
	}
  }

  if (m_pParent)
  {
      m_pParent->Command(m_Id);
  }
}

void CComboBox::Select(int item)
{
	gtk_list_select_item(GTK_LIST(GTK_COMBO(m_pWindow)->list),item);
	m_Num = item;
}

void CComboBox::ReadOnly(BOOL readOnly)
{
	gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(m_pWindow)->entry),!readOnly);
}

//------------------------------------------------------------------------------------------
//
//                       E N T R Y W I D G E T
//
//------------------------------------------------------------------------------------------
void CEntryWidget::CreateButton(PCTSTR title)
{
  m_pWindow = gtk_entry_new();
  if (title && (title[0]))
  {
    gtk_entry_set_text(GTK_ENTRY(m_pWindow),title);
  }
}

void CEntryWidget::SetText(PCTSTR text)
{
  gtk_entry_set_text(GTK_ENTRY(m_pWindow),text);
}

void CEntryWidget::AppendText(PCTSTR text)
{
  gtk_entry_append_text(GTK_ENTRY(m_pWindow),text);
}

void CEntryWidget::PrependText(PCTSTR text)
{
  gtk_entry_prepend_text(GTK_ENTRY(m_pWindow),text);
}

void CEntryWidget::Connect(void)
{
  gtk_signal_connect(GTK_OBJECT(m_pWindow),"changed",GTK_SIGNAL_FUNC(CS_EntryChanged),this);
}

gchar *CEntryWidget::GetText(void)
{
  gchar *ret=NULL;
  static char dummy[]="";

  if (m_pWindow)
  		ret = gtk_entry_get_text(GTK_ENTRY(m_pWindow));

  if (!ret) ret = dummy;
  return ret;
}

void CEntryWidget::SelectRegion(int nStart, int nEnd)
{
  gtk_entry_select_region(GTK_ENTRY(m_pWindow),nStart,nEnd);
}

void CEntryWidget::SetEditable(BOOL bEdit)
{
  gtk_entry_set_editable(GTK_ENTRY(m_pWindow),bEdit);
}

void CEntryWidget::SetMaxLength(int nMax)
{
  gtk_entry_set_max_length(GTK_ENTRY(m_pWindow),nMax);
}



//------------------------------------------------------------------------------------------
//
//                      R A D I O B U T T O N
//
//------------------------------------------------------------------------------------------

void CRadioButton::Init(void)
{	
//	CToggleButton::Init();
	m_pGroup=NULL;
}

void CRadioButton::CreateButton(PCTSTR title)
{
  m_pWindow = gtk_radio_button_new_with_label(m_pGroup,title);
}

int CRadioButton::Add(CRadioButton *pGroup, CWindow *pParent, int Id, PCTSTR title, BOOL show=TRUE)
{ 
      m_pGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(pGroup->m_pWindow));
      return CToggleButton::Create(pParent,Id,title,show);
}

int CRadioButton::Add(CRadioButton *pGroup, CPackBox *pParent, int Id, PCTSTR title, BOOL bExpand=FALSE, 
             BOOL bFill=FALSE, int nPadding=0, BOOL show=TRUE)
{
      m_pGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(pGroup->m_pWindow));
      return CToggleButton::Create(pParent,Id,title,bExpand,bFill,nPadding,show);
}

int CRadioButton::Add(CRadioButton *pGroup, CPackTable *pParent, int Id, 
             int nLeft, int nTop, int nRight, int nBottom, PCTSTR title,
             int nXopt=(GTK_FILL|GTK_EXPAND), int nYopt=(GTK_FILL|GTK_EXPAND), 
             int nXpadding=0, int nYpadding=0, BOOL show=TRUE)
{
      m_pGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(pGroup->m_pWindow));
      return CToggleButton::Create(pParent,Id,nLeft,nTop,nRight,nBottom,title,nXopt,nYopt,nXpadding,nYpadding,show);
}

