#include <gtk/gtk.h>

//----------------  Typdefinitionen ----------------------
typedef int BOOL;
typedef char *PSTR;
typedef const PSTR PCTSTR;

//--------------------------------------------------------

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define MB_OK 0x01
#define MB_NO 0x02
#define MB_YES 0x04
#define MB_CANCEL 0x08
#define MB_CONTINUE 0x10
#define MB_IGNORE 0x20

#define MB_SCREEN_CENTER 0x10000

//------------------ Debugausgaben ------------------------

void noprintf(PSTR,...);        // Dummy-Funktion (ansonsten wird eine WARNING angezeigt)
                                // Die Funktion befindet sich in MCAPP.CPP

#ifdef DEBUG
#define TRACE printf
#else
#define TRACE noprintf
#endif

//-----------------------------------------------------

class CSize
{
public:
	int width;
	int height;
	CSize(){width=0;height=0;}
	void operator=(const GtkRequisition&);
};

//--------------------------------------------------------   CSignal

gint CS_ButtonClick(GtkWidget *wi, gpointer klasse);
gint CS_SelectionChanged(GtkWidget *wi, gpointer klasse);
gint CS_ButtonEnter(GtkWidget *wi, gpointer klasse);
gint CS_ButtonLeave(GtkWidget *wi, gpointer klasse);
gint CS_ButtonPressed(GtkWidget *wi, gpointer klasse);
gint CS_ButtonPressed2(GtkWidget *wi, GdkEvent *event, gpointer klasse);
gint CS_ButtonReleased(GtkWidget *wi, gpointer klasse);
gint CS_ButtonReleased2(GtkWidget *wi, GdkEvent *event, gpointer klasse);
gint CS_ButtonToggled(GtkWidget *wi, gpointer klasse);
gint CS_EntryChanged(GtkWidget *wi, gpointer klasse);
gint CS_Destroy(GtkWidget *wi, GdkEvent *event, gpointer klasse);
gint CS_Quit(GtkWidget *wi, GdkEvent *event, gpointer klasse);

class CSignal
{
 public:
  virtual ~CSignal(){}
  static int m_DestroyAppRegistriert;
  
  virtual void Quit(void);
  virtual void Click(void);
  virtual void Enter(void);
  virtual void Leave(void);
  virtual void Press(void);
  virtual void Release(void);
  virtual void Toggle(void);
  virtual void Change(void);
  virtual void DestroyEvent(void);
  virtual void SelectionChanged(void);
};

//--------------------------------------------------------  CApp

class CApp : public CSignal
{
public:
  void Init(int *argc, char **argv[]);
  int Run(void);
};

//--------------------------------------------------------  CWidget

class CWindow;
class CWidget : public CSignal
{
  BOOL     m_Initialized;
 public:
  BOOL     m_StopCommandChain;   // FALSE -> "Command()" an Parent weiterleiten
  // Wird ein "Window" angelegt, dann wird diese Variable auf TRUE gesetzt
  // -> Weiterleitung stoppt beim ersten "richtigen" Fenster
  //    (Buttons, Boxes, usw. sind von CWindow abgeleitet, Nachrichten werden daher 
  //     weitergeleitet !)
  GtkWidget *m_pWindow;
  CWindow   *m_pParent;

 public:
  CWidget();

  void Connect(void);
  void Destroy(void);

  
  void Show(BOOL show=TRUE);
  int SetMinimumSize(int Width, int Height);
  CSize GetSize(void);
  void SetSize(const CSize &);
  void Enable(BOOL enable=TRUE);

  virtual BOOL Command(int Id);
  virtual void InitWidget(){;}
  virtual void CloseWidget(){;}
};

//-------------------------------------------------------  CWindow

class CWindow : public CWidget
{
public:
  BOOL   m_Modal;

  CWindow(){m_Modal=FALSE;}

  int    Create(GtkWindowType type, PCTSTR title, int Width, int Height, BOOL show=TRUE);
  void 	 SetTitle(PCTSTR title);
  int    SetDefaultSize(int Width, int Height);
  void   ContainerBorderWith(int border);
  void   SetModal(BOOL modal=TRUE);
  void   Center(void);
  int    MessageBox(PCTSTR text, PCTSTR title, int w=0, int h=0, unsigned long flags=MB_OK);
  void   EndModal(void);


  virtual void Quit(void);


};

//------------------------------------------------------  CPackBox
class CPackBox : public CWindow
{
 public:
  int Create(CWindow *pParent, BOOL bVertikal, BOOL bHomogen=TRUE, int nSpace=0, BOOL show=TRUE);
  int Create(CPackBox *pParent,BOOL bVertikal, BOOL bHomogen=TRUE, int nSpace=0, BOOL bExpand=FALSE,
             BOOL bFill=FALSE, int nPadding=0, BOOL show=TRUE);
  int CreateFromEnd(CPackBox *pParent,BOOL bVertikal, BOOL bHomogen=TRUE, int nSpace=0, BOOL bExpand=FALSE,
             BOOL bFill=FALSE, int nPadding=0, BOOL show=TRUE);
};
//------------------------------------------------------  CDialog
class CDialog : public CWindow
{
public:
   CPackBox m_ActionArea; //HBox (darf in "Destroy()" nicht zerstrt werden!)
   CPackBox m_Box; // VBox (darf in "Destroy()" nicht zerstrt werden!)
   int m_Ret;

   int Create(PCTSTR title, int w=0, int h=0, BOOL show=TRUE);
   int DoModal(void);
};
//------------------------------------------------------  CPackTable
class CPackTable : public CWindow
{
 public:
  int Create(CWindow *pParent, int nRows, int nCols, BOOL bHomogen=TRUE, BOOL show=TRUE);
  int Create(CPackBox *pParent, int nRows, int nCols, BOOL bHomogen=TRUE, int nSpace=0, 
	     BOOL bExpand=FALSE, BOOL bFill=FALSE, int nPadding=0, BOOL show=TRUE);
  int CreateFromEnd(CPackBox *pParent, int nRows, int nCols, BOOL bHomogen=TRUE, int nSpace=0, 
	     BOOL bExpand=FALSE, BOOL bFill=FALSE, int nPadding=0, BOOL show=TRUE);
};
//-------------------------------------------------------  CMenuBar
class CMenu;
class CMenuItem;
class CMenuBar : public CWidget 
{
public:
  int Create(CWindow *pParent, BOOL show=TRUE);
  int Create(CPackBox *pParent, BOOL bExpand=FALSE, BOOL bFill=FALSE, int nPadding=0, BOOL show=TRUE);
  int CreateFromEnd(CPackBox *pParent, BOOL bExpand=FALSE, BOOL bFill=FALSE, int nPadding=0, BOOL show=TRUE);
  int Create(CPackTable *pParent, int nLeft, int nTop, int nRight, int nBottom, PCTSTR title=NULL,
             int nXopt=(GTK_FILL|GTK_EXPAND), int nYopt=(GTK_FILL|GTK_EXPAND), 
             int nXpadding=0, int nYpadding=0, BOOL show=TRUE);
  void Append(CMenu *menu);
  void Append(CMenuItem *item);
};
//-------------------------------------------------------  CMenu
class CMenu
{
public:
   GtkWidget *m_pWindow;

   CMenu(){m_pWindow=NULL;}  // sobald von CMenu abgeleitet wird, mu das bliche Verfahren ber Init() verwendet werden!
   //~CMenu(){Destroy();}  // Destroy() nicht verwenden -> sonst gibts Fehlermeldungen
   //void Destroy(void){if (m_pWindow)gtk_widget_destroy(m_pWindow);m_pWindow=NULL;}

   int Create(void);
   void Append(CMenuItem *item);
};
//-------------------------------------------------------  CMenuItem
class CMenuItem : public CWidget
{
public:
   int m_Id;

   int Create(PCTSTR title, BOOL show=TRUE);
   int Create(CWindow *window, PCTSTR title, int id=-1, BOOL show=TRUE);
   void SetSubMenu(CMenu *menu);
   void Connect(void);
   
   virtual void Click(void);
};
//-------------------------------------------------------  CCheckMenuItem
class CCheckMenuItem : public CMenuItem
{
public:
	BOOL m_Checked;
   int Create(PCTSTR title, BOOL show=TRUE);
   int Create(CWindow *window, PCTSTR title, int id=-1, BOOL show=TRUE);
   void SetCheck(BOOL checked=TRUE);
   void Connect(void);


   virtual void Toggle(void);
};
//------------------------------------------------------  CScrolledWindow
class CScrolledWindow : public CWindow
{
public:
  int Create(CWindow *pParent, GtkAdjustment *horiz=NULL,GtkAdjustment *vert=NULL, BOOL show=TRUE);
  int Create(CPackBox *pParent, GtkAdjustment *horiz=NULL,GtkAdjustment *vert=NULL, BOOL bExpand=FALSE, BOOL bFill=FALSE, 
             int nPadding=0, BOOL show=TRUE);
  int CreateFromEnd(CPackBox *pParent, GtkAdjustment *horiz=NULL,GtkAdjustment *vert=NULL, BOOL bExpand=FALSE, BOOL bFill=FALSE, 
             int nPadding=0, BOOL show=TRUE);
  int Create(CPackTable *pParent, GtkAdjustment *horiz,GtkAdjustment *vert, int nLeft, int nTop, int nRight, int nBottom, PCTSTR title=NULL,
             int nXopt=(GTK_FILL|GTK_EXPAND), int nYopt=(GTK_FILL|GTK_EXPAND), 
             int nXpadding=0, int nYpadding=0, BOOL show=TRUE);
};
//------------------------------------------------------  CGenericButton
class CGenericButton : public CWindow
{
 private:
  virtual void CreateButton(PCTSTR titel) = 0;

 public:
  int m_Id;

  // Achtung bei nderungen von "Create()" evtl. auch CRadioButton anpassen !!!
  int Create(CWindow *pParent, int Id, PCTSTR title=NULL, BOOL show=TRUE);
  int Create(CPackBox *pParent, int Id, PCTSTR title=NULL, BOOL bExpand=FALSE, BOOL bFill=FALSE, 
             int nPadding=0, BOOL show=TRUE);
  int CreateFromEnd(CPackBox *pParent, int Id, PCTSTR title=NULL, BOOL bExpand=FALSE, BOOL bFill=FALSE, 
             int nPadding=0, BOOL show=TRUE);
  int Create(CPackTable *pParent, int Id, int nLeft, int nTop, int nRight, int nBottom, PCTSTR title=NULL,
             int nXopt=(GTK_FILL | GTK_EXPAND), int nYopt=(GTK_FILL | GTK_EXPAND), 
             int nXpadding=0, int nYpadding=0, BOOL show=TRUE);
  int Create(CScrolledWindow *pParent, int Id, PCTSTR title=NULL, BOOL show=TRUE);

  virtual void Connect(void);


  virtual void Enter(void);
  virtual void Leave(void);
};
//------------------------------------------------------  CLabel (in: mcbutton.cpp)
class CLabel : public CGenericButton
{
 private:
  virtual void CreateButton(PCTSTR titel);

 public:
  virtual void Connect(void);
  void SetText(PCTSTR text);
};
//------------------------------------------------------  CButton
class CButton : public CGenericButton
{
 private: 
  virtual void CreateButton(PCTSTR titel);

 public:
  CLabel m_pLabel;  // Darf nicht zerstrt werden !
 
  virtual void Connect(void);
  virtual void Click(void);
  virtual void Press(void);
  virtual void Release(void);
  void SetText(PCTSTR text);
};
//------------------------------------------------------  CDialog
class CMessageBox : public CDialog
{
public:
   CLabel m_Label;
   CButton m_Ok;
   CButton m_No;
   CButton m_Yes;
   CButton m_Cancel;
   CButton m_Continue;
   CButton m_Ignore;

   void Destroy(void);
   int Create(CWindow *parent, PCTSTR text, PCTSTR title, int w=0, int h=0, unsigned long flags=MB_OK);
   virtual BOOL Command(int Id);
};
//------------------------------------------------------  CList (in: mcbutton.cpp)
class CList : public CGenericButton
{
  gint m_nTitles;  // Anzahl der Titel
  gchar **m_pTitles;
  int m_nRows; // Anzahl Zeilen
 private:
  virtual void CreateButton(PCTSTR titel);

 public:
  CList();
  //CList(gint nTitles);
  //CList(gint nTitles, gchar *texte[]);
  ~CList(){TRACE("~CList -> fehlt\n");}

  void Init(void);
//  void Destroy(void){printf("~CList -> fehlt\n");}
  
//  int Create(CScrolledWindow *pParent, int Id, PCTSTR title=NULL, BOOL show=TRUE);
  int Create(CScrolledWindow *pParent, int Id, gint nTitles, gchar *texte[], BOOL show=TRUE);
  virtual void Connect(void);
  virtual void Click(void);

  void Clear(void);
  int GetRows(void);
  int Append(gchar *texte[]);
  int Prepend(gchar *texte[]);
  int Insert(gint row, gchar *texte[]);
  void Remove(gint row);
  void SetColumnWidth(gint column, gint width);
  gint GetOptimalColumnWidth(gint column);
  void ShowTitle(BOOL show=TRUE);
  void SetColumnActive(gint col, BOOL active=TRUE);
  void SetActive(BOOL active=TRUE);
};
//------------------------------------------------------  CListBox (in: mcbutton.cpp)
class CListBox : public CGenericButton
{
 private:
  virtual void CreateButton(PCTSTR titel);

 public:
  virtual void Connect(void);
  virtual void Click(void);

  void AddItem(PCTSTR title);
  void Clear(int nStart=0,int nEnd=-1);
};
//------------------------------------------------------  CComboBox (in: mcbutton.cpp)
class CComboBox : public CGenericButton
{
  GList *m_Liste;
 private:
  virtual void CreateButton(PCTSTR titel);

 public:
  int m_Num;
  
  CComboBox();
  ~CComboBox();

  void Destroy(void);
  virtual void Connect(void);
  virtual void Release(void);

  void AddItem(PCTSTR title);
  void Clear(int nStart,int nEnd);
  void Clear(void);
  void Select(int item);
  void ReadOnly(BOOL readOnly=TRUE);

};
//------------------------------------------------------  CEntryWidget (in: mcbutton.cpp)
class CEntryWidget : public CGenericButton
{
 private:
  virtual void CreateButton(PCTSTR titel);

 public:
  virtual void Connect(void);
  void   SetText(PCTSTR text);
  gchar  *GetText(void);
  void   AppendText(PCTSTR text);
  void   PrependText(PCTSTR text);
  void   SelectRegion(int nStart, int nEnd);
  void   SetEditable(BOOL bEdit);
  void   SetMaxLength(int nMax);
};
//------------------------------------------------------  CToggleButton (in: mcbutton.cpp)
class CToggleButton : public CButton
{
 private:
  virtual void CreateButton(PCTSTR titel);

 public:
  virtual void Connect(void);
  virtual void Toggle(void);
};
//------------------------------------------------------  CCheckButton (in: mcbutton.cpp)
class CCheckButton : public CToggleButton
{
 private:
  virtual void CreateButton(PCTSTR titel);

 public:
  void SetCheck(int markiert=1);
  int GetCheck(void);
};
//------------------------------------------------------  CRadioButton (in: mcbutton.cpp)
class CRadioButton : public CToggleButton
{
  GSList *m_pGroup;
 private:
  virtual void CreateButton(PCTSTR titel);

 public:
  CRadioButton(){Init();}
  ~CRadioButton(){printf("~CRadioButton -> fehlt\n");}

  void Init(void);
  void Destroy(void){printf("~CRadioButton -> fehlt\n");}
  
  int Add(CRadioButton *pGroup, CWindow *pParent, int Id, PCTSTR title, BOOL show=TRUE);
  int Add(CRadioButton *pGroup, CPackBox *pParent, int Id, PCTSTR title, BOOL bExpand=FALSE, 
             BOOL bFill=FALSE, int nPadding=0, BOOL show=TRUE);
  int Add(CRadioButton *pGroup, CPackTable *pParent, int Id,  
             int nLeft, int nTop, int nRight, int nBottom, PCTSTR title,
             int nXopt=(GTK_FILL|GTK_EXPAND), int nYopt=(GTK_FILL|GTK_EXPAND), 
             int nXpadding=0, int nYpadding=0, BOOL show=TRUE);
};
//-----------------------------------------------------

//-----------------------------------------------------

