#include <stdio.h>
#include "mcc.h"

int CWindow::Create(GtkWindowType type, PCTSTR title, int Width, int Height, BOOL show)
{

  m_StopCommandChain = TRUE;
  m_pWindow = gtk_window_new(type);
  if (!m_pWindow) return -1;

  SetTitle(title);
  SetDefaultSize(Width,Height);
  
  // Das erste TOP-Level-Window ist normalerweise immer das Hauptfenster
  if (type == GTK_WINDOW_TOPLEVEL)
  {
    if (!m_DestroyAppRegistriert) // FALSE => Hauptfenster
    {
	      m_DestroyAppRegistriert++;
	      m_Modal = TRUE;
    }
  }
  Connect();

  Show(show);   

  return 0;
}

int CWindow::MessageBox(PCTSTR text, PCTSTR title, int w=0, int h=0, unsigned long flags=MB_OK)
{
	CMessageBox msg;
	return msg.Create(this,text,title,w,h,flags);
}

void CWindow::EndModal(void)
{
	Destroy();
	gtk_main_quit();
}

void CWindow::Quit(void)
{
	//MessageBox("Test","Test");
	if(!m_Modal)
	{
		MessageBox("\n  Nichtmodale Fenster mssen  \n  durch Destroy() zerstrt werden!  \n","Debug-Hinweis");
		return;
	}
	EndModal();
}

int CWindow::SetDefaultSize(int Width, int Height)
{
  gtk_window_set_default_size(GTK_WINDOW(m_pWindow),Width,Height);
  return 0;
}

void CWindow::SetTitle(PCTSTR title)
{
	gtk_window_set_title(GTK_WINDOW(m_pWindow),title);
}

void CWindow::SetModal(BOOL modal)
{
  m_Modal = modal;
  gtk_window_set_modal(GTK_WINDOW(m_pWindow),modal);
}

void CWindow::ContainerBorderWith(int border)
{
  gtk_container_border_width(GTK_CONTAINER(m_pWindow),border);
}

void CWindow::Center(void)
{
	gtk_window_set_position(GTK_WINDOW(m_pWindow),GTK_WIN_POS_CENTER);
}

int CDialog::Create(PCTSTR title, int w, int h, BOOL show)
{
   m_Ret = 0;
   m_pParent = NULL;
   m_pWindow = gtk_dialog_new();
   if (!m_pWindow) return -1;
   if((w>0) && (h>0)) SetDefaultSize(w,h);
   m_Box.m_pParent = this;
   m_ActionArea.m_pParent = this;
   m_Box.m_pWindow = GTK_DIALOG(m_pWindow)->vbox;
   m_ActionArea.m_pWindow = GTK_DIALOG(m_pWindow)->action_area;
   SetTitle(title);
   Connect();
   Show(show);    
   return 0;
}

int CDialog::DoModal(void)
{
    SetModal();
    gtk_main();
    Destroy();
	return m_Ret;
}

void CMessageBox::Destroy(void)
{
   m_Label.Destroy();
   m_Ok.Destroy();
   m_No.Destroy();
   m_Yes.Destroy();
   m_Cancel.Destroy();
   m_Continue.Destroy();
   m_Ignore.Destroy();
}

int CMessageBox::Create(CWindow *parent, PCTSTR text, PCTSTR title, int w, int h, unsigned long flags)
{
   //printf("CMessageBox::Create %p\n",this);
   CDialog::Create(title,w,h,FALSE);
 
   m_Label.Create(&m_Box,-1,text,1,1);
   if (flags & MB_IGNORE)
   {
		m_Ignore.Create(&m_ActionArea,MB_IGNORE,"  Ignorieren ");
   }
   if (flags & MB_CONTINUE)
   {
   		m_Continue.Create(&m_ActionArea,MB_CONTINUE,"  Weiter  ");
   }
   if (flags & MB_CANCEL)
   {
   		m_Cancel.Create(&m_ActionArea,MB_CANCEL,"  Abbruch  ");
   }
   if (flags & MB_OK)
   {
		m_Ok.Create(&m_ActionArea,MB_OK,"  Ok  ");
   }
   if (flags & MB_NO)
   {
		m_No.Create(&m_ActionArea,MB_NO,"  Nein  ");
   }
   if (flags & MB_YES)
   {
		m_Yes.Create(&m_ActionArea,MB_YES,"  Ja  ");
   }

   if ((flags & MB_SCREEN_CENTER) || !parent)
   		gtk_window_set_position(GTK_WINDOW(m_pWindow),GTK_WIN_POS_CENTER);
   else
   {
		//gint x,y;
		//gtk_widget_get_pointer(parent->m_pWindow,&x,&y);
   		gtk_window_set_position(GTK_WINDOW(m_pWindow),GTK_WIN_POS_CENTER);
   		//printf("x %d y %d\n",x,y);
   		//gtk_widget_set_uposition(m_pWindow,x,y);
   }


   Show();
   int ret = DoModal();
   Destroy();
   return ret;
}

BOOL CMessageBox::Command(int Id)
{
    m_Ret = Id;
    gtk_main_quit();
    //Destroy();
    //printf("-2-\n");
    return TRUE;
}

int CScrolledWindow::Create(CWindow *pParent, GtkAdjustment *horiz,GtkAdjustment *vert, BOOL show)
{
    m_pParent = pParent;
    m_pWindow = gtk_scrolled_window_new(horiz,vert);
    if (!m_pWindow) return -1;
    if (m_pParent) gtk_container_add(GTK_CONTAINER(m_pParent->m_pWindow),m_pWindow);
  	Show(show);    
    return 0;
}

int CScrolledWindow::Create(CPackBox *pParent, GtkAdjustment *horiz,GtkAdjustment *vert, BOOL bExpand, BOOL bFill, int nPadding, BOOL show)
{
  m_pParent = pParent;
  m_pWindow = gtk_scrolled_window_new(horiz,vert);
  if (!m_pWindow) return -1;
  if (m_pParent) gtk_box_pack_start(GTK_BOX(m_pParent->m_pWindow),m_pWindow,bExpand,bFill,nPadding);
 // Connect();
  Show(show);    
  return 0;
}

int CScrolledWindow::CreateFromEnd(CPackBox *pParent, GtkAdjustment *horiz,GtkAdjustment *vert, BOOL bExpand, BOOL bFill, int nPadding, BOOL show)
{
  m_pParent = pParent;
  m_pWindow = gtk_scrolled_window_new(horiz,vert);
  if (!m_pWindow) return -1;
  if (m_pParent) gtk_box_pack_end(GTK_BOX(m_pParent->m_pWindow),m_pWindow,bExpand,bFill,nPadding);
 // Connect();
  Show(show);    
  return 0;
}

int CScrolledWindow::Create(CPackTable *pParent, GtkAdjustment *horiz,GtkAdjustment *vert, int nLeft, int nTop, int nRight, int nBottom, PCTSTR title, int nXopt, int nYopt, int nXpadding, int nYpadding, BOOL show)
{
  m_pParent = pParent;
  m_pWindow = gtk_scrolled_window_new(horiz,vert);
  if (!m_pWindow) return -1;
  if (m_pParent) 
  {
    gtk_table_attach(GTK_TABLE(pParent->m_pWindow),m_pWindow,nLeft,nRight,nTop,nBottom,(GtkAttachOptions)nXopt,(GtkAttachOptions)nYopt,nXpadding,nYpadding);
  }
  Show(show);    
  return 0;
}


