#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <fcntl.h>
#include <unistd.h>    // wegen "sleep()"
#include <errno.h>    // fr Zugriffe auf "errno"
#include "mcc.h"
#include "test1.h"

CApp theApp;
MyWindow mainwindow;
char *titel[] = {"Action","Time","Id","Rtr","Dlc","Message","Lost messages"};
char test[][50] = {"Transmitted","0:00","0123","1","2","02 33","0"};
char *zz[10];
int thread_loop=0;
pthread_t thread1;
struct ISRInfo thread_transmit;
int input_id=0;
int file=-1;
int procpipe[2];
int procpipe2[2];
  

int TransmitToCAN(struct ISRInfo *isr); 
void ResetTime(); 
int ConvertString(unsigned char *ziel, int max, const char * quelle);


struct PipeCommand
{
	int intern;
	union 
	{
		struct ISRInfo info;
		struct CanParInfo cpi;
		struct opencan
		{
			char name[50];
			int type;
			int bit;
		} oc;
		struct wrongversion
		{
			int ist;
			int soll;
		} wv;
		struct canaccess
		{
			unsigned long mask;
			unsigned long code;
		} ca;
		struct threaddata
		{
			struct ISRInfo info;
			unsigned long time;
		}td;
		unsigned long baudrate;
		//unsigned long time;
		//unsigned long mask;
		//unsigned long code;
	};
};

int enum {
	CANBUS_RECEIVED_DATA=0,
	CANBUS_TRANSMIT_DATA,
	CANBUS_QUIT_PROCESS,
	CANBUS_STANDBY,
	CANBUS_CLEAR_OVERRUN,
	CANBUS_ABORT_TRANSMISSION,
	CANBUS_RESET_TIME,
	CANBUS_CANT_OPEN_DRIVER,
	CANBUS_WRONG_VERSION,
	CANBUS_OPEN,
	CANBUS_CLOSE,
	CANBUS_ACCESS,
	CANBUS_BAUDRATE,
	CANBUS_THREAD_START,
	CANBUS_THREAD_STOP,
};

int DeviceIOControl(int file, int cmd, void *inBuf, int inCount, void *outBuf, int outCount, int *dwRet) 
{
    int ret;
    
    if (dwRet)
    	*dwRet = 0;

	if (file == -1)
		return 0;
		
  	struct ioctlTransfer trans;
	memset(&trans,0,sizeof(trans));

	trans.cbInBuffer = inCount;
	trans.lpvInBuffer = inBuf;
	trans.cbOutBuffer = outCount;
	trans.lpvOutBuffer = outBuf;
	ret = ioctl(file, cmd, &trans);
    if (dwRet)
    	*dwRet = trans.cbBytesReturned;

	return ret;
}


HIGHLOW Div64To32(HIGHLOW hl, unsigned long n)
{
	unsigned long h,l;
	HIGHLOW ret;

	h = hl.high & 0xffff; // Verwende nur 16 BIT der Sekunden (es wird nur die Differenz bentigt)
	l = hl.low / 1000;	// Auflsung ndern in ms (Originalwert hat us)

	ret.high = 0;
	ret.low = h*1000+l;

	ret.low -= n;

	return ret;
}



void *thread(void *ptr)
{
	struct timespec ts2, tsr2;
	struct timespec ts, tsr;
    int ret;
    unsigned long time = (unsigned long)ptr;
    int errors=0;

	ts.tv_sec  = time/1000;
	ts.tv_nsec = (time%1000)*1000000;

	ts2.tv_sec  = 300/1000;
	ts2.tv_nsec = (300%1000)*1000000;
	//TRACE("Thread %ld %ld %ld\n",mw->m_thread_time,ts.tv_sec,ts.tv_nsec);

    while(TRUE)
    {
    	if (!thread_loop)
    	{
    		break;
    	}
   		ret = TransmitToCAN(&thread_transmit);
   		if(!ret)
   		{
   			errors++;
   			if(errors > 10)
   				break;
   			//sleep(1);
   			nanosleep (&ts2, &tsr2);
   		}
   		else
   		{
   			errors=0;
   		}
		if(time)
    		nanosleep (&ts, &tsr);
    }

	thread_loop = 0;
    return 0;
}

void Refresh(int signum)
{
	struct PipeCommand pi;
	int dwRet;

	// Das Signal wurde im Treiber gesetzt -> Hole jetzt die Daten ab
	do
	{
		// Daten werden in einer Schleife solange abgeholt, bis:
		//  - keine Daten mehr vorhanden sind (info.isrNummer <> 0)
		//  - ein Fehler bei DeviceIOControl() auftrat
		//  - DeviceIOControl() keine (d.h. auch keine INFO-Struktur) zurckgab
		if(DeviceIOControl(file, IOCTL_IO_13, 0, 0, &pi.info, sizeof(struct ISRInfo), &dwRet) < 0)
		{
			break;
		}

		if (!dwRet)
		{
			break;
		}
		pi.intern = CANBUS_RECEIVED_DATA;
		write (procpipe[1],&pi, sizeof(pi));
	}while(pi.info.isrNummer);
	return;
}

void Init4CAN(void)
{
	struct sigaction sa;
	memset(&sa,0,sizeof(sa));
	sa.sa_handler = Refresh;
	sa.sa_flags = SA_RESTART;
	sigaction(SIGIO,&sa,0);
}

void DeInit4CAN(void)
{
	if(file != -1)
	{
		close(file);
	}
}

void StartThread(struct PipeCommand *p)
{
	if(!thread_loop)
	{
		thread_loop = 1;
		memcpy(&thread_transmit,&p->td.info,sizeof(struct ISRInfo));
		pthread_create(&thread1,NULL,thread,(void*)p->td.time);
		pthread_detach(thread1);
	}
}

void StopThread(void)
{
 	if(thread_loop)
	{
		thread_loop = 0;
	}
}

int CANClose() 
{
	if(file != -1)
	{
		close(file);
		file = -1;
	}
	return 1;
}

int CANOpen(struct PipeCommand *in) 
{
	struct CanParInfo info;
	struct PipeCommand p;
	if(file != -1)
		close(file);

	file = open(in->oc.name,O_RDWR);
	//TRACE("Ergebnis Treiber ffnen: %d\n",file);
	if(file == -1)
	{
		TRACE("Kann Treiber nicht ffnen\n");
		p.intern = CANBUS_CANT_OPEN_DRIVER;
		if (write(procpipe[1],&p,sizeof(p)) < 0)
		{
			fprintf(stderr,"Can't write data to pipe");
			perror(0);
		}
		return 0;
	}

	fcntl(file,F_SETOWN,getpid());
	fcntl(file,F_SETFL, FASYNC | fcntl(file,F_GETFL));
	
	memset(&info,0,sizeof(info));
	DeviceIOControl(file, IOCTL_IO_1, 0, 0, &info, sizeof(info), 0); 
	//TRACE("%x %x\n",IOCTL_IO_1,IOCTL_IO_4);
	TRACE("Nach ioctl: V%ld Card %d OS %d\n",info.version,info.card,info.os);
	if (info.version != CANPAR_VERSION)
	{
		TRACE("Falsche Treiber Version\n");
		close(file);
		file = -1;
		p.intern = CANBUS_WRONG_VERSION;
		p.wv.ist = info.version;
		p.wv.soll = CANPAR_VERSION;
		if (write(procpipe[1],&p,sizeof(p)) < 0)
		{
			fprintf(stderr,"Can't write data to pipe");
			perror(0);
		}
		return 0;
	}

    char tmp;
	tmp = in->oc.type;// 0=Basic Can  1=PeliCan
	TRACE("2.0A/2.0B: %d\n",tmp);
	DeviceIOControl(file, IOCTL_IO_14, &tmp, 1, 0, 0, 0); 

	tmp = in->oc.bit;// 0=SFF 1=EFF
	TRACE("11/29: %d\n",tmp);
	DeviceIOControl(file, IOCTL_IO_15, &tmp, 1, 0, 0, 0); 
	ResetTime();
	return 1;
}

void GoToSleep() 
{
	char tmp;
	tmp = CAN_COMMAND_STANDBY;
	TRACE("GoToSleep\n");
	DeviceIOControl(file, IOCTL_IO_16, &tmp, 1, 0, 0, 0); 
}

void ClearDataOverrun() 
{
	char tmp;
	tmp = CAN_COMMAND_CLEAR_OVERRUN;
	DeviceIOControl(file, IOCTL_IO_16, &tmp, 1, 0, 0, 0); 
}

void AbortTransmission() 
{
	char tmp;
	tmp = CAN_COMMAND_ABORT_TRANSMISSION;
	DeviceIOControl(file, IOCTL_IO_16, &tmp, 1, 0, 0, 0); 
}

void ResetTime() 
{
	struct PipeCommand p;
	p.intern = CANBUS_RESET_TIME;
	DeviceIOControl(file, IOCTL_IO_1, 0, 0, &p.cpi, sizeof(struct CanParInfo), 0); 
	if (write(procpipe[1],&p,sizeof(p)) < 0)
	{
		fprintf(stderr,"Can't write data to pipe");
		perror(0);
	}
}

void OnBaudrateChanged(struct PipeCommand *p) 
{
	DeviceIOControl(file, IOCTL_IO_11, &p->baudrate, sizeof(unsigned long), 0, 0, 0); 
}

void AcceptanceCode_Mask(struct PipeCommand *p) 
{
	// Um den Acceptance Code in das SJA1000 zu schreiben
	// wird das ResetRequest-Bit gesetzt. Dadurch wird eine
	// laufende Botschaft abgebrochen.
	unsigned long tmp[2];
	tmp[0] = p->ca.code;
	tmp[1] = p->ca.mask;
	DeviceIOControl(file, IOCTL_IO_9, tmp, sizeof(tmp), 0, 0, 0); 
}

int TransmitToCAN(struct ISRInfo *isr) 
{
	if(DeviceIOControl(file, IOCTL_IO_10, isr, sizeof(ISRInfo), 0, 0, 0) < 0)
	{
		return 0;
	}
	return 1;
}

int  MyWindow::CANClose() 
{
	struct PipeCommand p;
	p.intern = CANBUS_CLOSE;
	if (write(procpipe2[1],&p,sizeof(p)) < 0)
	{
		fprintf(stderr,"Can't write data to pipe");
		perror(0);
		return 0;
	}
	return 1;
}

int  MyWindow::CANOpen() 
{
	struct PipeCommand p;

	Geraet dlg;
 			
	dlg.Create("CAN Treiber ffnen",100,50);
	dlg.DoModal();

	p.intern = CANBUS_OPEN;
	strcpy(p.oc.name,"/dev/canpar");
	p.oc.type = dlg.typ.m_Num;// 0=Basic Can  1=PeliCan
	p.oc.bit = dlg.identifier.m_Num;// 0=SFF 1=EFF
	if (write(procpipe2[1],&p,sizeof(p)) < 0)
	{
		fprintf(stderr,"Can't write data to pipe");
		perror(0);
		return 0;
	}

	OnAcceptanceCode_Mask();
	OnBaudrateChanged();

	return 1;
}

void MyWindow::StartThread(void)
{
	long t;
	sscanf(m_Botschaften.zeit.GetText(),"%ld",&t);

	struct PipeCommand p;
	p.intern = CANBUS_THREAD_START;
	p.td.time = t;
	memcpy(&p.td.info,&thread_transmit,sizeof(struct ISRInfo));
	if (write(procpipe2[1],&p,sizeof(p)) < 0)
	{
		fprintf(stderr,"Can't write data to pipe");
		perror(0);
	}
}

void MyWindow::StopThread(void)
{
	struct PipeCommand p;
	p.intern = CANBUS_THREAD_STOP;
	if (write(procpipe2[1],&p,sizeof(p)) < 0)
	{
		fprintf(stderr,"Can't write data to pipe");
		perror(0);
	}
}

void MyWindow::OnBaudrateChanged() 
{
	struct PipeCommand p;
	p.intern = CANBUS_BAUDRATE;
	p.baudrate = m_Baudrate;
	if (write(procpipe2[1],&p,sizeof(p)) < 0)
	{
		fprintf(stderr,"Can't write data to pipe");
		perror(0);
	}
}

void MyWindow::OnAcceptanceCode_Mask() 
{
	struct PipeCommand p;
	p.intern = CANBUS_ACCESS;
	p.ca.code = m_Code;
	p.ca.mask = m_Mask;
	if (write(procpipe2[1],&p,sizeof(p)) < 0)
	{
		fprintf(stderr,"Can't write data to pipe");
		perror(0);
	}
}

int MyWindow::Transmit(long id, short rtr, short dlc, unsigned char *data) 
{
	struct PipeCommand p;
	int t;

	p.intern = CANBUS_TRANSMIT_DATA;
	p.info.canDLC = (unsigned char)dlc;
	p.info.canIdentifier = id;
	p.info.canRTR = (unsigned char)rtr;
	for(t=0;t<dlc;t++)
		p.info.canMsg[t] = (unsigned char)data[t];

	if (write(procpipe2[1],&p,sizeof(p)) < 0)
	{
		fprintf(stderr,"Can't write data to pipe");
		perror(0);
		return 0;
	}
	return 1;
}

int MyWindow::ResetTime() 
{
	struct PipeCommand p;
	p.intern = CANBUS_RESET_TIME;
	if (write(procpipe2[1],&p,sizeof(p)) < 0)
	{
		fprintf(stderr,"Can't write data to pipe");
		perror(0);
		return 0;
	}
	return 1;
}

int MyWindow::GoToSleep() 
{
	struct PipeCommand p;
	p.intern = CANBUS_STANDBY;
	if (write(procpipe2[1],&p,sizeof(p)) < 0)
	{
		fprintf(stderr,"Can't write data to pipe");
		perror(0);
		return 0;
	}
	return 1;
}

int MyWindow::ClearDataOverrun() 
{
	struct PipeCommand p;
	p.intern = CANBUS_CLEAR_OVERRUN;
	if (write(procpipe2[1],&p,sizeof(p)) < 0)
	{
		fprintf(stderr,"Can't write data to pipe");
		perror(0);
		return 0;
	}
	return 1;
}

int MyWindow::AbortTransmission() 
{
	struct PipeCommand p;
	p.intern = CANBUS_ABORT_TRANSMISSION;
	if (write(procpipe2[1],&p,sizeof(p)) < 0)
	{
		fprintf(stderr,"Can't write data to pipe");
		perror(0);
		return 0;
	}
	return 1;
}

void StopWorkerProcess() 
{
	struct PipeCommand p;
	p.intern = CANBUS_QUIT_PROCESS;
	write(procpipe2[1],&p,sizeof(p));
}



MyWindow::MyWindow()
{

    zz[0] = test[0];
    zz[1] = test[1];
    zz[2] = test[2];
    zz[3] = test[3];
    zz[4] = test[4];
    zz[5] = test[5];
    zz[6] = test[6];
    m_Mask=0xffffffff;
    m_Code=0xffffffff;
    m_ShowMessages = TRUE;

    m_oldTime=0;		// Zur Berechnung der Differenz zwischen zwei Botschaften
 	m_startTime=0;	// Startzeit (wird in OpenCan() festgestellt)
 	//ready = 0;
}

void MyWindow::AddDataToListe(void)
{
    	b8.Prepend(zz);
    	if(b8.GetRows() >= 50)
    	{
    	   b8.Remove(49);
    	}
    	gdk_flush();
}

BOOL MyWindow::Command(int Id)
{
    switch (Id)
    {
    default:
    	return FALSE;
    case IDM_QUIT:
    	Quit();
    	break;
    case 3:
    	break;
	case IDM_RESET_TIME:
		ResetTime();
		break;
	case IDM_SHOW_DATA:
		TRACE("IDM_SHOW_DATA wurde aufgerufen %d\n",m_ShowMessages);
		m_ShowMessages = !m_ShowMessages;
    	break;
	case IDM_GO_TO_SLEEP:
		GoToSleep();
    	break;
	case IDM_CLEAR_OVR:
		ClearDataOverrun();
    	break;
	case IDM_ABORT_TRANSMISION:
		AbortTransmission();
		break;
    case IDM_ACCEPTANCE_CODE:
    	{
    		CanAccess dlg;
    		dlg.m_Mask = m_Mask;
    		dlg.m_Code = m_Code;
    		dlg.Create("Acceptance Code + Mask",100,50);
    		dlg.DoModal();
    		m_Mask = dlg.m_Mask;
    		m_Code = dlg.m_Code;
			OnAcceptanceCode_Mask();
    	}
    	break;
    case IDM_DRV_OPEN:   // Device ffnen
    	CANOpen();
    	break;
    case IDM_DRV_CLOSE:   // Device schlieen
    	CANClose();
		break;
    case IDM_BAUDRATE:   // Baudrate
    	{
    		Baudrate dlg;
    		dlg.m_Wert = m_Baudrate;
   			dlg.Create("Baudrate",100,50);
    		dlg.DoModal();
    		m_Baudrate = dlg.m_Wert;
    		OnBaudrateChanged();
   		}
    	break;
    case IDM_MESSAGES:	// Botschaften eingeben
    	m_Botschaften.Show();
    	break;
    case IDM_INFO:
    	{
    		CMessageBox msg;
    		msg.Create(this,"   Test for CANPAR-driver   \n 2001 Juergen Eder","Info about...",200,100,MB_OK);
    	}
    	break;
    }
    return TRUE;
}


int ConvertString(unsigned char *ziel, int max, const char * quelle)
{
   // String in Block wandeln
   int  z = 0;
   int  t = 0;
   int  x;
   int  w_temp;
   char temp[10];
   BOOL start=TRUE;

   for(x=0;x<=(int)strlen(quelle);x++)
   {
		 if ((quelle[x] != ' ') && quelle[x])
		 {
			 temp[z] = quelle[x];
			 z++;
			 start = FALSE;
		 }
		 if (!start && ((z > 1) || (quelle[x] == ' ') || !quelle[x]))
		 {
			 temp[z] = 0;
			 z=0;
			 start = TRUE;
			 sscanf(temp,"%x",&w_temp);
			 // Umgewandeltes Datenbyte
			 ziel[t] = (unsigned char)w_temp;
			 t++;
			 if (t >= max) break;
		 }
   }
   return t;
}


BOOL MyWindow::TransmitData(long id, short rtr, const char * data) 
{
    unsigned char x[8];
	short m;
	BOOL ret;
	m = (short)ConvertString(x,8,data);
	ret = Transmit(id,rtr,m,x);
	return ret;
}

void MyWindow::xFireCANError(struct ISRInfo *msg)
{
	if (msg->intern == 1)
	{
		TRACE("CAN-Karte wurde rckgesetzt (Spannung aus/ein)\n");
		OnAcceptanceCode_Mask();
		OnBaudrateChanged();
	}
	else
	{
		HIGHLOW x;
		sprintf(test[0],"Error");
	    if (!m_ShowMessages) return;
		x = Div64To32(msg->timeStamp,m_startTime);
		sprintf(test[1],"%.3f s  [%.3f s]",(double)x.low/1000.,(double)(x.low-m_oldTime)/1000.); m_oldTime = x.low;
		test[2][0]=0;
		test[3][0]=0;
		test[4][0]=0;
		test[5][0]=0;
		AddDataToListe();
	}

}

void MyWindow::xFireCANWakeUp(struct ISRInfo *msg)
{
	if (!m_ShowMessages) return;
	sprintf(test[0],"WakeUp");
	HIGHLOW x;
	x = Div64To32(msg->timeStamp,m_startTime);
	sprintf(test[1],"%.3f s  [%.3f s]",(double)x.low/1000.,(double)(x.low-m_oldTime)/1000.); m_oldTime = x.low;
	test[2][0]=0;
	test[3][0]=0;
	test[4][0]=0;
	test[5][0]=0;
	AddDataToListe();
}

void MyWindow::xFireCANOverrun(struct ISRInfo *msg)
{
	if (!m_ShowMessages) return;
	sprintf(test[0],"Overrun");
	HIGHLOW x;
	x = Div64To32(msg->timeStamp,m_startTime);
	sprintf(test[1],"%.3f s  [%.3f s]",(double)x.low/1000.,(double)(x.low-m_oldTime)/1000.); m_oldTime = x.low;
	test[2][0]=0;
	test[3][0]=0;
	test[4][0]=0;
	test[5][0]=0;
	AddDataToListe();
}

void MyWindow::xFireCANReceived(struct ISRInfo *msg)
{

	if (msg->lost)
	{
		m_DataLost += msg->lost;
	}
	if (!m_ShowMessages) return;
	sprintf(test[0],"Received");
	HIGHLOW x;
	x = Div64To32(msg->timeStamp,m_startTime);
	sprintf(test[1],"%.3f s  [%.3f s]",(double)x.low/1000.,(double)(x.low-m_oldTime)/1000.); m_oldTime = x.low;
	//sprintf(test[1],"-:-- s [-:--- s]");
	sprintf(test[2],"%lx",msg->canIdentifier);
	sprintf(test[3],"%d",msg->canRTR);
	sprintf(test[4],"%d",msg->canDLC);
	test[5][0]=0;
	for(int tmp=0;tmp<msg->canDLC;tmp++)
		sprintf(&test[5][strlen(test[5])],"%02x ",msg->canMsg[tmp]);
	sprintf(test[6],"%ld",m_DataLost);
	AddDataToListe();
}

void MyWindow::xFireCANTransmitted(struct ISRInfo *msg)
{
	if (!m_ShowMessages) return;

	sprintf(test[0],"Transmitted");
	HIGHLOW x;
	x = Div64To32(msg->timeStamp,m_startTime);
	sprintf(test[1],"%.3f s  [%.3f s]",(double)x.low/1000.,(double)(x.low-m_oldTime)/1000.); m_oldTime = x.low;
	sprintf(test[2],"%lx",msg->canIdentifier);
	sprintf(test[3],"%d",msg->canRTR);
	sprintf(test[4],"%d",msg->canDLC);
	test[5][0]=0;
	for(int tmp=0;tmp<msg->canDLC;tmp++)
		sprintf(&test[5][strlen(test[5])],"%02x ",msg->canMsg[tmp]);
	test[6][0] = 0;
	AddDataToListe();
}

void MyWindow::InitWidget()
{
	//TRACE("MyWindow::InitWindow\n");
	m_Baudrate = 20000;

	m_Botschaften.Create("CAN messages",300,200,FALSE);
	m_Botschaften.hauptfenster = this;
	
  	mainwindow.ContainerBorderWith(5);
  	box.Create(&mainwindow,TRUE,FALSE,0);

  	mn.Create(&box,FALSE,FALSE,2);
  	mdatei.Create();
  	mconfig.Create();
  	mhilfe.Create();

 	 datei.Create("File");
      	beenden.Create(&mainwindow,"Quit",IDM_QUIT);

  	config.Create("Configuration");
		oeffnen.Create(&mainwindow,"Open CAN...",IDM_DRV_OPEN);
		schliessen.Create(&mainwindow,"Close CAN",IDM_DRV_CLOSE);

		baudrate.Create(&mainwindow,"Baudrate...",IDM_BAUDRATE);
		acceptance.Create(&mainwindow,"Acceptance filter...",IDM_ACCEPTANCE_CODE);
		botschaften.Create(&mainwindow,"Transmit messages...",IDM_MESSAGES);

		reset_time.Create(&mainwindow,"Reset time",IDM_RESET_TIME);
		show_data.Create(&mainwindow,"Show data",IDM_SHOW_DATA);

		go_to_sleep.Create(&mainwindow,"Go to sleep",IDM_GO_TO_SLEEP);
		clear_ovr.Create(&mainwindow,"Clear data overrun",IDM_CLEAR_OVR);
		abort_trans.Create(&mainwindow,"Abort transmission",IDM_ABORT_TRANSMISION);

  	hilfe.Create("Help");
  		info.Create(&mainwindow,"Info about...",IDM_INFO);

  	datei.SetSubMenu(&mdatei);
  	config.SetSubMenu(&mconfig);
  	hilfe.SetSubMenu(&mhilfe);

  	mdatei.Append(&beenden);

  	mconfig.Append(&oeffnen);
  	mconfig.Append(&schliessen);
  	mconfig.Append(&baudrate);
	mconfig.Append(&acceptance);
  	mconfig.Append(&botschaften);
  	mconfig.Append(&reset_time);
  	mconfig.Append(&show_data);
  	mconfig.Append(&go_to_sleep);
  	mconfig.Append(&clear_ovr);
  	mconfig.Append(&abort_trans);

 	 mhilfe.Append(&info);

  	mn.Append(&datei);
  	mn.Append(&config);
  	mn.Append(&hilfe);

  	show_data.SetCheck();
  	
  	table.CreateFromEnd(&box,11,2,TRUE,0,TRUE,TRUE);
  	scrwin.Create(&table,NULL,NULL,0,0,2,11);

  	//b3.Create(&table,3,1,0,2,1,"Add Item");

  	b8.Create(&scrwin, 8, 7, titel);
  	b8.SetColumnWidth(0,80);
  	b8.SetColumnWidth(1,80);
  	b8.SetColumnWidth(2,60);
  	b8.SetColumnWidth(3,30);
  	b8.SetColumnWidth(4,30);
  	b8.SetColumnWidth(5,120);
  	b8.SetActive(FALSE);

	FILE *in;
    char buffer[100];
    sprintf(buffer,"%s/.catt.ini",g_get_home_dir());
	in = fopen(buffer,"r");
	while(in)
	{
		if(!fgets(buffer,sizeof(buffer),in))break;
		m_Baudrate = atoi(buffer);

		if(!fgets(buffer,sizeof(buffer),in))break;
		buffer[strlen(buffer)-1] =  0;
		m_Botschaften.c1.SetText(buffer);
		if(!fgets(buffer,sizeof(buffer),in))break;
		buffer[strlen(buffer)-1] =  0;
		m_Botschaften.c2.SetText(buffer);
		if(!fgets(buffer,sizeof(buffer),in))break;
		buffer[strlen(buffer)-1] =  0;
		m_Botschaften.c3.SetText(buffer);
		if(!fgets(buffer,sizeof(buffer),in))break;
		buffer[strlen(buffer)-1] =  0;
		m_Botschaften.c4.SetText(buffer);
		if(!fgets(buffer,sizeof(buffer),in))break;
		buffer[strlen(buffer)-1] =  0;
		m_Botschaften.c5.SetText(buffer);

		if(!fgets(buffer,sizeof(buffer),in))break;
		buffer[strlen(buffer)-1] =  0;
		m_Botschaften.b1.SetText(buffer);
		if(!fgets(buffer,sizeof(buffer),in))break;
		buffer[strlen(buffer)-1] =  0;
		m_Botschaften.b2.SetText(buffer);
		if(!fgets(buffer,sizeof(buffer),in))break;
		buffer[strlen(buffer)-1] =  0;
		m_Botschaften.b3.SetText(buffer);
		if(!fgets(buffer,sizeof(buffer),in))break;
		buffer[strlen(buffer)-1] =  0;
		m_Botschaften.b4.SetText(buffer);
		if(!fgets(buffer,sizeof(buffer),in))break;
		buffer[strlen(buffer)-1] =  0;
		m_Botschaften.b5.SetText(buffer);
		
		if(!fgets(buffer,sizeof(buffer),in))break;
		buffer[strlen(buffer)-1] =  0;
		m_Botschaften.zeit.SetText(buffer);

		buffer[0] = 0;
		if(!fgets(buffer,sizeof(buffer),in))break;
		m_Botschaften.s1.SetCheck(atoi(buffer));
		if(!fgets(buffer,sizeof(buffer),in))break;
		m_Botschaften.s2.SetCheck(atoi(buffer));
		if(!fgets(buffer,sizeof(buffer),in))break;
		m_Botschaften.s3.SetCheck(atoi(buffer));
		if(!fgets(buffer,sizeof(buffer),in))break;
		m_Botschaften.s4.SetCheck(atoi(buffer));

		if(!fgets(buffer,sizeof(buffer),in))break;
		m_Mask = atol(buffer);
		if(!fgets(buffer,sizeof(buffer),in))break;
		m_Code = atol(buffer);

		fclose(in);
		break;
	}
	//ready = 1;
}

void MyWindow::CloseWidget()
{
	TRACE("MyWindow::CloseWidget\n");
	FILE *in;
    char buffer[100];

    StopThread();
    gdk_input_remove(input_id);
    //ready = 0;
	TRACE("MyWindow::CloseWidget -2- \n");
    //thread_loop = 0;

	b3.Destroy();
	b8.Destroy();
	table.Destroy();
	box.Destroy();
	scrwin.Destroy();
    
    sprintf(buffer,"%s/.catt.ini",g_get_home_dir());
    //TRACE("WriteConfig %s\n",buffer);
	in = fopen(buffer,"w");
	if(in)
	{
		fprintf(in,"%ld\n",m_Baudrate);

		fprintf(in,"%s\n",m_Botschaften.c1.GetText());
		fprintf(in,"%s\n",m_Botschaften.c2.GetText());
		fprintf(in,"%s\n",m_Botschaften.c3.GetText());
		fprintf(in,"%s\n",m_Botschaften.c4.GetText());
		fprintf(in,"%s\n",m_Botschaften.c5.GetText());

		fprintf(in,"%s\n",m_Botschaften.b1.GetText());
		fprintf(in,"%s\n",m_Botschaften.b2.GetText());
		fprintf(in,"%s\n",m_Botschaften.b3.GetText());
		fprintf(in,"%s\n",m_Botschaften.b4.GetText());
		fprintf(in,"%s\n",m_Botschaften.b5.GetText());

		fprintf(in,"%s\n",m_Botschaften.zeit.GetText());

		fprintf(in,"%d\n",m_Botschaften.s1.GetCheck());
		fprintf(in,"%d\n",m_Botschaften.s2.GetCheck());
		fprintf(in,"%d\n",m_Botschaften.s3.GetCheck());
		fprintf(in,"%d\n",m_Botschaften.s4.GetCheck());

		fprintf(in,"%lu\n",m_Mask);
		fprintf(in,"%lu\n",m_Code);

		fclose(in);
	}
	else
	{
		TRACE("Kann Konfiguration (%s)nicht schreiben\n",buffer);
	}
	m_Botschaften.Destroy();
}

void CanAccess::InitWidget()
{
	char buffer[20];
	p1.Create(&m_Box,FALSE);
	p2.Create(&m_Box,FALSE);
   	Center();
   	l1.Create(&p1,-1,"Acceptance Code");
   	sprintf(buffer,"%lx",m_Code);
   	code.Create(&p1,2,buffer);
   	l2.Create(&p2,-1,"Acceptance Mask");
   	sprintf(buffer,"%lx",m_Mask);
   	mask.Create(&p2,1,buffer);
	ok.Create(&m_ActionArea,100,"  Ok  ");
}

void CanAccess::CloseWidget()
{
	unsigned long t;
	sscanf(mask.GetText(),"%lx",&t); m_Mask = t;
	sscanf(code.GetText(),"%lx",&t); m_Code = t;
	mask.Destroy();
	code.Destroy();
	ok.Destroy();
}

BOOL CanAccess::Command(int Id)
{
    switch(Id)
    {
    	case 100:				// OK
    		EndModal();
    		break;
    }
    return TRUE;
}

void Geraet::InitWidget()
{
   	Center();

	ok.Create(&m_ActionArea,100,"  Ok  ");

	liste.Create(&m_Box,1,"Devices",1,1);
	liste.AddItem("/dev/CANPAR");
	liste.AddItem("/dev/CANPAR1");
	liste.AddItem("/dev/CANPAR2");
	liste.AddItem("/dev/CANPAR3");
	liste.Select(num);
	liste.ReadOnly();

	typ.Create(&m_Box,2,"Type",1,1);
	typ.AddItem("CAN 2.0 A");
	typ.AddItem("CAN 2.0 B");
	typ.Select(0);
	typ.ReadOnly();
	
	identifier.Create(&m_Box,3,"Identifier",1,1);
	identifier.AddItem("11 Bit");
	identifier.AddItem("29 Bit");
	identifier.Select(0);
	identifier.ReadOnly();
	identifier.Enable(FALSE);

	identifier.Enable(FALSE);
}

void Geraet::CloseWidget()
{
	liste.Destroy();
	typ.Destroy();
	identifier.Destroy();
	ok.Destroy();
}

BOOL Geraet::Command(int Id)
{
    switch(Id)
    {
    	case 1:
    		num = liste.m_Num;
	   		//TRACE("Num %d\n",num);
    		break;
    	case 2:
    		if (typ.m_Num > 0)
    			identifier.Enable();
    		else
    			identifier.Enable(FALSE);
    		break;
    	case 100:				// OK
    		m_Ret = Id;
    		EndModal();
    		break;
    }
    return TRUE;
}

void Baudrate::InitWidget()
{
	char buffer[20];
	//TRACE("InitWidget %ld\n",m_Wert);
	sprintf(buffer,"%ld",m_Wert);
   	baud.Create(&m_Box,1,buffer,1,1);
   	ok.Create(&m_ActionArea,2,"  Ok  ");
    Center();
	//baud.SetText("123");
}

void Baudrate::CloseWidget()
{
	sscanf(baud.GetText(),"%ld",&m_Wert);
	baud.Destroy();
	ok.Destroy();
}

BOOL Baudrate::Command(int Id)
{
    switch(Id)
    {
    default:
    	//TRACE("Baudrate::Command %d\n",Id);
    	break;
    case 2:				// OK
    	gtk_main_quit();
    	Destroy();
    	break;
    }
    return TRUE;
}

void Botschaften::InitWidget()
{
	table.Create(&m_Box,6,10,0);

   	CSize size;
    int p=0;
   	s1.Create(&table,50,p,1,p+1,2);
   	s2.Create(&table,51,p,2,p+1,3);
   	s3.Create(&table,52,p,3,p+1,4);
   	s4.Create(&table,53,p,4,p+1,5);

	p += 1;
   	l1.Create(&table,1,p,0,p+1,1,"Can ID");
   	c1.Create(&table,10,p,1,p+1,2);
   	size = c1.GetSize();
   	size.width = 50;
   	c1.SetSize(size);
   	c2.Create(&table,11,p,2,p+1,3);
   	c2.SetSize(size);
   	c3.Create(&table,12,p,3,p+1,4);
   	c3.SetSize(size);
   	c4.Create(&table,13,p,4,p+1,5);
   	c4.SetSize(size);
   	c5.Create(&table,14,p,5,p+1,6,"",(GTK_FILL|GTK_EXPAND),(GTK_FILL|GTK_EXPAND),0,10,TRUE);
   	c5.SetSize(size);

    p += 1;
   	l2.Create(&table,2,p,0,p+4,1,"CAN message");
   	b1.Create(&table,20,p,1,p+4,2);
   	b2.Create(&table,21,p,2,p+4,3);
   	b3.Create(&table,22,p,3,p+4,4);
   	b4.Create(&table,23,p,4,p+4,5);
   	b5.Create(&table,24,p,5,p+4,6);

    p += 4;
   	l3.Create(&table,3,p,0,p+1,1,"Time [ms]");
   	zeit.Create(&table,30,p,5,p+1,6,"0");
   	zeit.SetSize(size);

   	p += 1;
   	l4.Create(&table,4,p,0,p+1,1,"Action");
   	senden.Create(&table,40,p,1,p+1,2," Send ");
   	start.Create(&table,42,p,5,p+1,6," Start ",(GTK_FILL|GTK_EXPAND),(GTK_FILL|GTK_EXPAND),2,10,TRUE);

   	p += 1;
   	abbrechen.Create(&table,41,p,1,p+1,2," Cancel ");
   	stop.Create(&table,43,p,5,p+1,6," Stop ",(GTK_FILL|GTK_EXPAND),(GTK_FILL|GTK_EXPAND),2,10,TRUE);

   	
   	ok.Create(&m_ActionArea,100,"  Close  ");
    Center();
}

void Botschaften::CloseWidget()
{
    //MessageBox("CloseWidget","Test");
	table.Destroy();
	b1.Destroy();
	b2.Destroy();
	b3.Destroy();
	b4.Destroy();
	b5.Destroy();
	c1.Destroy();
	c2.Destroy();
	c3.Destroy();
	c4.Destroy();
	c5.Destroy();
	l1.Destroy();
	l2.Destroy();
	l3.Destroy();
	l4.Destroy();
	l5.Destroy();
	zeit.Destroy();
	senden.Destroy();
	abbrechen.Destroy();
	start.Destroy();
	stop.Destroy();
	s1.Destroy();
	s2.Destroy();
	s3.Destroy();
	s4.Destroy();
	ok.Destroy();
}

BOOL Botschaften::SendeDatenZeileX(CEntryWidget *id, CEntryWidget *dat, CCheckButton *mark)
{
    if(mark && !mark->GetCheck())
    	return 1;

	unsigned long i;
	sscanf(id->GetText(),"%lx",&i);
	return hauptfenster->TransmitData(i,0,dat->GetText());
}

BOOL Botschaften::SendeDaten(void)
{
	while(1)
	{
		if(!SendeDatenZeileX(&c1,&b1,&s1)) break;
		if(!SendeDatenZeileX(&c2,&b2,&s2)) break;
		if(!SendeDatenZeileX(&c3,&b3,&s3)) break;
		if(!SendeDatenZeileX(&c4,&b4,&s4)) break;
		break;
	}
	return TRUE;
}

BOOL Botschaften::SendDatenAusThread()
{
	unsigned long i;
    unsigned char x[8];
	short m;
	int t;
	sscanf(c5.GetText(),"%lx",&i);
	m = (short)ConvertString(x,8,b5.GetText());
	thread_transmit.canDLC = (unsigned char)m;
	thread_transmit.canIdentifier = i;
	thread_transmit.canRTR = 0;
	for(t=0;t<m;t++)
		thread_transmit.canMsg[t] = (unsigned char)x[t];
	return TRUE;
}

BOOL Botschaften::Command(int Id)
{
    switch(Id)
    {
    default:
    	TRACE("Baudrate::Command %d\n",Id);
    	break;
    case 40:	// Senden
    	SendeDaten();
    	break;
    case 41:	// Abbrechen
    	hauptfenster->AbortTransmission();
    	break;
    case 42: 	// Start
    	SendDatenAusThread();
    	hauptfenster->StartThread();
    	break;
    case 43:	// Stop
    	hauptfenster->StopThread();
    	break;
   	case 100:				// OK
   		Show(FALSE);
   		break;
    }
    return TRUE;
}

void DatenLesenVonBus(gpointer data, gint source,GdkInputCondition condition)
{
	struct PipeCommand pi;

	read(procpipe[0],&pi, sizeof(pi));
	TRACE("EMpfange Nachricht\n");
	switch(pi.intern) // Nachricht vom Treiber
	{
		case 0:
			if (pi.info.isrNummer)
			{
				switch (pi.info.msg)
				{
				case EV2PRG_RECEIVED:
					mainwindow.xFireCANReceived(&pi.info);
					break;
				case EV2PRG_TRANSMIT:
					mainwindow.xFireCANTransmitted(&pi.info);
					break;
				case EV2PRG_ERROR:
				case EV2PRG_WARNING:
				case EV2PRG_PASSIVE:
				case EV2PRG_ARBITRATION:
				case EV2PRG_BUSERROR:
					mainwindow.xFireCANError(&pi.info);
					break;
				case EV2PRG_WAKEUP:
					mainwindow.xFireCANWakeUp(&pi.info);
				case EV2PRG_OVERRUN:
					mainwindow.xFireCANOverrun(&pi.info);
					break;
				}
			}
			break;
		case CANBUS_RESET_TIME:
			TRACE("Reset Time\n");
			HIGHLOW x = Div64To32(pi.cpi.time,0);
			mainwindow.m_oldTime = 0;
			mainwindow.m_startTime = x.low;
			break;
		case CANBUS_WRONG_VERSION:
			{
				char buffer[100];
				sprintf(buffer,"Wrong driver version: %d",pi.wv.ist);
				mainwindow.MessageBox(buffer,"Error",200,100,MB_OK);
			}
			break;
		case CANBUS_CANT_OPEN_DRIVER:
			mainwindow.MessageBox("Can't open driver","Error",200,100,MB_OK);
			break;

	}
}

int main(int argc, char *argv[])
{
  
  theApp.Init(&argc,&argv);
  mainwindow.Create(GTK_WINDOW_TOPLEVEL,"C.A.T.T. for Linux",600,300);

  // Pipe vom "Arbeitsproze" zum "Viewproze"
  if (pipe(procpipe))
  {
      fprintf (stderr, "Can't create Pipe\n");
      return EXIT_FAILURE;
  }  

  // Pipe vom "Viewproze" zum "Arbeitsproze"
  if (pipe(procpipe2))
  {
      fprintf (stderr, "Can't create Pipe\n");
      return EXIT_FAILURE;
  }  

  int ret = fork();
  if (ret < 0)
  {
  	fprintf(stderr, "Can't create process, terminating\n");
  	return EXIT_FAILURE;
  }
  if (ret == 0)
  {
  	// Bin jetzt im erzeugten Proze
  	int loop=1;
	struct PipeCommand pi;

	close(procpipe2[1]);
  	Init4CAN();

	while(loop)
	{
  		pi.intern = -1;	
        read(procpipe2[0],&pi, sizeof(pi));
		TRACE("Empfange Befehl %d\n",pi.intern);
		switch(pi.intern)
		{
		case CANBUS_TRANSMIT_DATA:
			TransmitToCAN(&pi.info); 
			break;
		case CANBUS_QUIT_PROCESS:
			loop=0;
			break;
		case CANBUS_STANDBY:
			GoToSleep();
			break;
		case CANBUS_CLEAR_OVERRUN:
			ClearDataOverrun();
			break;
		case CANBUS_ABORT_TRANSMISSION:
			AbortTransmission();
			break;
		case CANBUS_RESET_TIME:
			ResetTime();
			break;
		case CANBUS_OPEN:
			CANOpen(&pi);
			break;
		case CANBUS_CLOSE:
			CANClose();
			break;
		case CANBUS_BAUDRATE:
			OnBaudrateChanged(&pi);
			break;
		case CANBUS_ACCESS:
			AcceptanceCode_Mask(&pi);
			break;
		case CANBUS_THREAD_START:
			StartThread(&pi);
			break;
		case CANBUS_THREAD_STOP:
			StopThread();
			break;
		}
	}

  	StopThread();
  	DeInit4CAN();
	close(procpipe2[0]);
	sleep(1);
  	return 0;
  }
  if (ret > 0)
  {
	close(procpipe2[0]);
    input_id = gdk_input_add(procpipe[0],GDK_INPUT_READ,DatenLesenVonBus,0);
  	theApp.Run();  
  	StopWorkerProcess();
	close(procpipe2[1]);
  }
  return 0;
}

